"""
Screenshot and video export functionalities for vtk viewer.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Callable, Tuple
from os.path import splitext

import imageio
import numpy as np
from PIL import Image
from qtpy.QtGui import QImage, QGuiApplication
from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QSpinBox,
    QComboBox,
    QPushButton,
    QGroupBox,
    QRadioButton,
    QFileDialog,
    QMessageBox,
    QGridLayout,
    QApplication,
)
from vtkmodules.util import numpy_support
from vtkmodules.vtkRenderingCore import vtkWindowToImageFilter


__all__ = ["ExportManager"]


def _get_trajectories(geometries):
    from .geometry import GeometryTrajectory

    return [x for x in geometries if isinstance(x, GeometryTrajectory)]


class FrameWriter:
    def __init__(self, filename):
        self.index = 0
        self.filename, self.ext = splitext(filename)

    def append_data(self, img: np.ndarray):
        image = Image.fromarray(np.asarray(img))
        image.save(f"{self.filename}_{self.index}{self.ext}")
        self.index += 1

    def release(self):
        self.index = 0


class ExportManager:
    def __init__(self, vtk_widget, volume_viewer, cdata):
        self.vtk_widget = vtk_widget
        self.viewer = volume_viewer
        self.cdata = cdata

        self.format_settings = {
            "MP4": {"ext": ".mp4"},
            "AVI": {"ext": ".avi"},
            "RGBA": {"frame_series": True, "ext": ".png"},
        }

    def copy_screenshot_to_clipboard(self, window: bool = False):
        screenshot = self.capture_screenshot(transparent_bg=True)
        if window:
            screenshot = self.capture_window_screenshot(transparent_bg=True)

        screenshot = np.array(screenshot)
        height, width = screenshot.shape[:2]

        q_image = QImage(
            screenshot.data,
            width,
            height,
            width * 4,
            QImage.Format.Format_RGBA8888,
        )

        clipboard = QGuiApplication.clipboard()
        return clipboard.setImage(q_image)

    def capture_screenshot(
        self,
        transparent_bg: bool = False,
        width: int = None,
        height: int = None,
        magnification: int = 2,
        multisamples: int = 8,
    ):
        """Capture high-quality screenshot of current VTK window

        Parameters
        ----------
        transparent_bg : bool, optional
            Whether to keep transparent background, by default False
        width : int, optional
            Custom width for the screenshot, uses current window width by default.
        height : int, optional
            Custom height for the screenshot, uses current window height by default.
        magnification : int, optional
            Resolution multiplier for supersampling (1-8), by default 2
        multisamples : int, optional
            Number of multisamples for hardware antialiasing, by default 8

        Returns
        -------
        PIL.Image
            Screenshot image
        """
        render_window = self.vtk_widget.GetRenderWindow()
        render_window.SetAlphaBitPlanes(1)

        original_multisamples = render_window.GetMultiSamples()
        original_size = render_window.GetSize()

        render_window.SetMultiSamples(multisamples)

        size_changed = False
        target_width = width if width is not None else original_size[0]
        target_height = height if height is not None else original_size[1]

        # Apply magnification for supersampling
        render_width = target_width * magnification
        render_height = target_height * magnification

        if render_width != original_size[0] or render_height != original_size[1]:
            render_window.SetSize(render_width, render_height)
            render_window.Render()
            size_changed = True

        window_to_image = vtkWindowToImageFilter()
        window_to_image.SetInput(render_window)
        window_to_image.SetInputBufferTypeToRGBA()
        window_to_image.SetScale(1)
        window_to_image.ReadFrontBufferOff()
        window_to_image.Update()

        vtk_image = window_to_image.GetOutput()
        img_width, img_height, _ = vtk_image.GetDimensions()

        arr = numpy_support.vtk_to_numpy(vtk_image.GetPointData().GetScalars())
        arr = arr.reshape(img_height, img_width, -1)[::-1]

        render_window.SetMultiSamples(original_multisamples)

        if size_changed:
            render_window.SetSize(*original_size)
            render_window.Render()

        ret = Image.fromarray(arr, "RGBA")

        if magnification > 1:
            ret = ret.resize((target_width, target_height), Image.LANCZOS)

        if transparent_bg:
            return ret
        return ret.convert("RGB")

    def capture_window_screenshot(self, transparent_bg: bool = False):
        """Capture a screenshot of the entire PyQt window application."""
        top_window = self.vtk_widget
        while top_window.parent():
            top_window = top_window.parent()

        geometry = top_window.geometry()
        screen = QGuiApplication.screenAt(top_window.mapToGlobal(geometry.topLeft()))
        if screen is None:
            raise RuntimeError("Could not find screen containing the window")

        image = screen.grabWindow(
            top_window.winId(), 0, 0, geometry.width(), geometry.height()
        ).toImage()

        width = image.width()
        height = image.height()
        ptr = image.constBits()
        ptr.setsize(height * width * 4)

        arr = np.frombuffer(ptr, np.uint8).reshape(height, width, -1).copy()
        arr[:, :, [0, 2]] = arr[:, :, [2, 0]]
        ret = Image.fromarray(arr, "RGBA")
        if transparent_bg:
            return ret
        return ret.convert("RGB")

    def save_screenshot(self):
        file_path, _ = QFileDialog.getSaveFileName(
            None, "Save Screenshot", "", "Images (*.png *.jpg)"
        )
        if not file_path:
            return -1

        transparent_bg = file_path.lower().endswith(".png")
        screenshot = self.capture_screenshot(transparent_bg=transparent_bg)
        screenshot.save(file_path)

    def export_animation(self):
        has_volume = getattr(self.viewer.primary, "volume", None) is not None
        has_trajectory = len(_get_trajectories(self.cdata._models.data))

        dialog = AnimationSettingsDialog(
            self.viewer, self.cdata, self.format_settings.keys()
        )

        if not dialog.exec():
            return -1

        use_reveal = dialog.reveal_radio.isChecked()
        use_trajectory = dialog.trajectory_radio.isChecked()

        if dialog.start_frame.value() > dialog.end_frame.value():
            QMessageBox.warning(
                None,
                "Export Error",
                "Start frame larger than stop frame.",
            )
            return -1

        if not (has_trajectory or has_volume):
            QMessageBox.warning(
                None,
                "Export Error",
                "Please load a trajectory or volume to animate.",
            )
            return -1

        actors = (
            self.vtk_widget.GetRenderWindow()
            .GetRenderers()
            .GetFirstRenderer()
            .GetActors()
        )
        actors_visible = []
        actors.InitTraversal()
        for _ in range(actors.GetNumberOfItems()):
            actor = actors.GetNextActor()
            if actor:
                actors_visible.append(actor.GetVisibility())

        original_frame = 0
        if not use_trajectory:
            original_frame = self.viewer.primary.slice_slider.value()

        stride = dialog.frame_stride.value()
        start, end = dialog.start_frame.value(), dialog.end_frame.value()
        frames = tuple(range(start, end, stride))
        if use_reveal:
            frames = (*frames, None, *list(range(end, start, -stride)))
            # self.viewer.primary.project_selector.setCurrentText("Project -")
            actors.InitTraversal()
            for _ in range(actors.GetNumberOfItems()):
                actors.GetNextActor().SetVisibility(0)

        def _update(frame_idx):
            if use_trajectory:
                for geometry in _get_trajectories(self.cdata._models.data):
                    geometry.display_frame(frame_idx)

                uuids = self.cdata.models._get_selected_uuids()
                if uuids:
                    return self.cdata.models.set_selection_by_uuids(uuids)
                return self.cdata.models.render_vtk()

            if use_reveal and frame_idx is None:
                # self.viewer.primary.project_selector.setCurrentText("Off")
                actors.InitTraversal()
                for visible in actors_visible:
                    actors.GetNextActor().SetVisibility(visible)
                return None

            if frame_idx is not None:
                return self.viewer.primary.slice_slider.setValue(frame_idx)

        try:
            self.vtk_widget.GetRenderWindow().SetOffScreenRendering(1)
            self._create_animation(
                update_func=_update,
                original_frame=original_frame,
                fps=dialog.frame_rate.value(),
                frames=frames,
                format_settings=self.format_settings[dialog.format_combo.currentText()],
            )
        except Exception as e:
            QMessageBox.warning(None, "Error", str(e))
        finally:
            self.vtk_widget.GetRenderWindow().SetOffScreenRendering(0)

    def _create_animation(
        self,
        update_func: Callable,
        fps: int,
        frames: Tuple[int],
        format_settings: dict,
        original_frame: int = None,
        quality: int = 80,
    ):
        """Create animation video from frames"""
        from .dialogs import ProgressDialog

        filename, _ = QFileDialog.getSaveFileName(
            None, "Save Animation", "", f"Video (*{format_settings['ext']})"
        )
        if filename is None or len(filename) == 0 or len(frames) == 0:
            return -1

        renderer = self.vtk_widget.GetRenderWindow()

        is_video = not format_settings.get("frame_series", False)

        update_func(frames[0])
        renderer.Render()
        screenshot = self.capture_screenshot(not is_video)
        height, width = np.array(screenshot).shape[:2]

        if is_video:
            height += height % 2
            width += width % 2

        screenshot_kwargs = {
            "transparent_bg": not is_video,
            "width": width,
            "height": height,
            "magnification": 1,
        }
        writer = FrameWriter(filename=filename)
        quality = max(min(quality / 10.0, 10), 1) if quality else None
        if is_video:
            writer = imageio.get_writer(
                filename, mode="I", fps=fps, quality=quality, macro_block_size=None
            )

        with ProgressDialog(frames, title="Processing Frames") as progress:
            for frame_idx in progress:
                update_func(frame_idx)
                if frame_idx is None:
                    continue
                renderer.Render()
                frame = np.asarray(self.capture_screenshot(**screenshot_kwargs))
                writer.append_data(frame)
                QApplication.processEvents()

            writer.close()

        if original_frame is not None:
            update_func(original_frame)


class AnimationSettingsDialog(QDialog):
    def __init__(self, volume_viewer, cdata, formats, parent=None):
        from .stylesheets import QGroupBox_style, QPushButton_style

        super().__init__(parent)
        self.viewer = volume_viewer
        self.cdata = cdata
        self.formats = formats
        self.setWindowTitle("Animation Settings")
        self.setup_ui()
        self.setStyleSheet(QGroupBox_style + QPushButton_style)

    def setup_ui(self):
        from .icons import dialog_accept_icon, dialog_reject_icon

        layout = QVBoxLayout(self)

        type_group = QGroupBox("Animation Type")
        type_layout = QVBoxLayout()
        self.trajectory_radio = QRadioButton("Trajectory")
        self.slice_radio = QRadioButton("Slices")
        self.reveal_radio = QRadioButton("Reveal Flythrough")

        has_volume = getattr(self.viewer.primary, "volume", None) is not None
        has_trajectory = len(_get_trajectories(self.cdata._models.data))

        self.slice_radio.setEnabled(has_volume)
        self.reveal_radio.setEnabled(has_volume)
        self.trajectory_radio.setEnabled(has_trajectory)
        if has_trajectory:
            self.trajectory_radio.setChecked(True)
        elif has_volume:
            self.slice_radio.setChecked(True)

        self.trajectory_radio.toggled.connect(self.update_frame_ranges)
        self.slice_radio.toggled.connect(self.update_frame_ranges)
        self.reveal_radio.toggled.connect(self.update_frame_ranges)
        type_layout.addWidget(self.trajectory_radio)
        type_layout.addWidget(self.slice_radio)
        type_layout.addWidget(self.reveal_radio)
        type_group.setLayout(type_layout)
        layout.addWidget(type_group)

        export_format_group = QGroupBox("Export Settings")
        export_format_grid = QGridLayout()
        export_format_grid.addWidget(QLabel("Format:"), 0, 0)
        self.format_combo = QComboBox()
        self.format_combo.addItems(self.formats)
        export_format_grid.addWidget(self.format_combo, 0, 1)
        export_format_grid.addWidget(QLabel("Quality:"), 1, 0)
        self.quality = QSpinBox()
        self.quality.setRange(0, 100)
        self.quality.setValue(80)
        self.quality.setSuffix("%")
        export_format_grid.addWidget(self.quality, 1, 1)
        export_format_group.setLayout(export_format_grid)
        layout.addWidget(export_format_group)

        # Frame settings
        frame_group = QGroupBox("Frame Settings")
        frame_grid = QGridLayout()

        frame_grid.addWidget(QLabel("Rate (fps):"), 0, 0)
        self.frame_rate = QSpinBox()
        self.frame_rate.setRange(1, 1 << 30)
        self.frame_rate.setValue(30)
        frame_grid.addWidget(self.frame_rate, 0, 1)

        frame_grid.addWidget(QLabel("Stride:"), 1, 0)
        self.frame_stride = QSpinBox()
        self.frame_stride.setRange(1, 1 << 30)
        self.frame_stride.setValue(1)
        frame_grid.addWidget(self.frame_stride, 1, 1)

        frame_grid.addWidget(QLabel("Window:"), 2, 0)
        range_layout = QHBoxLayout()
        self.start_frame = QSpinBox()
        self.start_frame.setFixedWidth(70)
        self.end_frame = QSpinBox()
        self.end_frame.setFixedWidth(70)
        range_layout.addWidget(self.start_frame)
        range_layout.addWidget(self.end_frame)
        frame_grid.addLayout(range_layout, 2, 1)

        frame_group.setLayout(frame_grid)
        layout.addWidget(frame_group)

        button_layout = QHBoxLayout()
        ok_button = QPushButton("Done")
        ok_button.clicked.connect(self.accept)
        ok_button.setIcon(dialog_accept_icon)

        cancel_button = QPushButton("Cancel")
        cancel_button.clicked.connect(self.reject)
        cancel_button.setIcon(dialog_reject_icon)

        button_layout.addWidget(cancel_button)
        button_layout.addWidget(ok_button)
        layout.addLayout(button_layout)

        self.update_frame_ranges()

    def update_frame_ranges(self):
        min_frame, max_frame = 0, 0
        if self.trajectory_radio.isChecked():
            max_frame = max(
                *[x.frames - 1 for x in _get_trajectories(self.cdata._models.data)], 0
            )
        elif self.slice_radio.isChecked() or self.reveal_radio.isChecked():
            if self.viewer.primary.volume is not None:
                max_frame = self.viewer.primary.slice_slider.maximum()
                min_frame = self.viewer.primary.slice_slider.minimum()

        self.start_frame.setRange(min_frame, max_frame)
        self.end_frame.setRange(min_frame, max_frame)
        self.start_frame.setValue(min_frame)
        self.end_frame.setValue(max_frame)



from qtpy.QtWidgets import QDialog, QVBoxLayout, QGroupBox

from ..widgets import HistogramWidget


class HistogramDialog(QDialog):
    def __init__(self, cdata, parent=None):
        super().__init__(parent)
        self.cdata = cdata

        layout = QVBoxLayout(self)
        group = QGroupBox("Select")
        group_layout = QVBoxLayout(group)

        self.histogram_widget = HistogramWidget()
        group_layout.addWidget(self.histogram_widget)
        layout.addWidget(group)

        self.cdata.data.render_update.connect(self.update_histogram)
        self.histogram_widget.cutoff_changed.connect(self._on_cutoff_changed)
        self.update_histogram()

    def get_cluster_size(self):
        return [x.get_number_of_points() for x in self.cdata._data.data]

    def update_histogram(self, data=None):
        self.histogram_widget.update_histogram(self.get_cluster_size())

    def _on_cutoff_changed(self, lower_cutoff, upper_cutoff=None):
        cluster_sizes = self.get_cluster_size()
        if upper_cutoff is None:
            upper_cutoff = max(cluster_sizes) + 1

        uuids = []
        for geometry in self.cdata._data.data:
            n_points = geometry.get_number_of_points()
            if (n_points > lower_cutoff) & (n_points < upper_cutoff):
                uuids.append(geometry.uuid)
        self.cdata.data.set_selection_by_uuid(uuids)

    def closeEvent(self, event):
        """Disconnect when dialog closes"""
        try:
            self.cdata.data.render_update.disconnect(self.update_histogram)
            self.histogram_widget.cutoff_changed.disconnect(self._on_cutoff_changed)
        except (TypeError, RuntimeError):
            pass  # Already disconnected
        super().closeEvent(event)



from typing import Dict
from qtpy.QtWidgets import QFileDialog, QComboBox
from qtpy.QtCore import Signal


class MappedComboBox(QComboBox):
    customPathSelected = Signal(str)

    def __init__(self, parent=None, choices: Dict = {}):
        super().__init__(parent)

        self.setup_models(choices)
        self.currentTextChanged.connect(self._handle_selection)
        self.setSizePolicy(QComboBox().sizePolicy())

    def setup_models(self, choices):
        self.clear()
        self.predefined_models = choices | {"Browse...": None}
        self.addItems(self.predefined_models.keys())

    def _handle_selection(self, text):
        if text == "Browse...":
            file_path, _ = QFileDialog.getOpenFileName(
                self,
                "Select Model Checkpoint",
                "",
                "Checkpoint Files (*.ckpt *.pth);;All Files (*.*)",
            )

            if file_path:
                display_name = file_path
                self.predefined_models[display_name] = file_path

                self.blockSignals(True)
                self.insertItem(self.count() - 1, display_name)
                self.setCurrentText(display_name)
                self.blockSignals(False)

                self.customPathSelected.emit(file_path)
            else:
                self.blockSignals(True)
                self.setCurrentIndex(0)
                self.blockSignals(False)

    def get_selected_path(self):
        return self.predefine



from qtpy.QtWidgets import QWidget, QSlider, QDoubleSpinBox, QVBoxLayout
from qtpy.QtCore import Qt, Signal


class SliderWithInput(QWidget):
    valueChanged = Signal(int)

    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super().__init__(parent)
        self.setup_ui(orientation)

    def setup_ui(self, orientation):
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.slider = QSlider(orientation)
        self.spinbox = QDoubleSpinBox()
        self.spinbox.setRange(0, 100)
        self.spinbox.setDecimals(4)
        self.spinbox.setSingleStep(0.1)

        self.slider.valueChanged.connect(self.valueChanged.emit)
        self.slider.valueChanged.connect(self._update_spinbox_from_slider)
        self.spinbox.valueChanged.connect(self._update_slider_from_spinbox)

        main_layout.addWidget(self.slider)
        main_layout.addWidget(self.spinbox)

        self.setLayout(main_layout)

    def _update_spinbox_from_slider(self):
        self.spinbox.blockSignals(True)
        self.spinbox.setValue(self._value_to_quantile(self.slider.value()))
        self.spinbox.blockSignals(False)

    def _update_slider_from_spinbox(self):
        self.slider.setValue(self._quantile_to_value(self.spinbox.value()))

    def _quantile_to_value(self, value) -> int:
        return int(value * (self.maximum() - self.minimum()) / 100 + self.minimum())

    def _value_to_quantile(self, value) -> float:
        return 100 * (value - self.minimum()) / (self.maximum() - self.minimum())

    def setRange(self, minimum, maximum):
        self.slider.setRange(minimum, maximum)

    def setValue(self, value):
        self.slider.setValue(value)

    def value(self):
        return self.slider.value()

    def setOrientation(self, orientation):
        self.slider.setOrientation(orientation)

    def minimum(self):
        return self.slider.minimum()

    def maximum(self):
        return self.slider.maximum()



import re
from os import listdir
from typing import Union
from platform import system
from os.path import join, exists, basename

import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout, QFileDialog, QMessageBox, QApplication

from ..parallel import submit_task
from ..widgets.ribbon import create_button
from ..stylesheets import QPushButton_style


def on_run_complete(self, *args, **kwargs):
    self.cdata.data.render()
    self.cdata.models.render()


def _getExistingDirectory(parent, text):
    dialog = QFileDialog(parent)
    dialog.setWindowTitle(text)

    dialog.setFileMode(QFileDialog.Directory)

    # The native dialog on macOS omits the dialog text which can be confusing
    if system() == "Darwin":
        dialog.setOptions(
            QFileDialog.DontUseCustomDirectoryIcons | QFileDialog.DontUseNativeDialog
        )
    else:
        dialog.setOptions(QFileDialog.DontUseCustomDirectoryIcons)

    dialog.setStyleSheet(QPushButton_style)
    return dialog


class IntelligenceTab(QWidget):
    def __init__(self, cdata, ribbon, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon

        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

    def show_ribbon(self):
        from ..segmentation import MEMBRAIN_SETTINGS
        from ..dialogs import TemplateMatchingDialog

        self.ribbon.clear()

        hmff_actions = [
            create_button("Equilibrate", "mdi.molecule", self, self._equilibrate_fit),
            create_button("Setup", "mdi.export", self, self._setup_hmff),
            create_button(
                "Trajectory",
                "mdi.chart-line-variant",
                self,
                self._import_trajectory,
                "Import Trajectory",
                IMPORT_SETTINGS,
            ),
            create_button("Backmapping", "mdi.set-merge", self, self._map_fit),
        ]
        self.ribbon.add_section("DTS Simulation", hmff_actions)

        matching_actions = [
            create_button(
                "Setup",
                "mdi.magnify",
                self,
                lambda: TemplateMatchingDialog().exec_(),
                "Identify proteins using template matching",
            ),
        ]
        self.ribbon.add_section("Template Matching", matching_actions)

        segmentation_actions = [
            create_button("Add", "mdi.plus", self, self.add_cloud, "Add test data"),
            create_button(
                "Membrane",
                "mdi.border-all-variant",
                self,
                self._run_membrane_segmentation,
                "Segment membranes using Membrain-seg",
                MEMBRAIN_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Segmentation", segmentation_actions)

    def _equilibrate_fit(self):
        from ..dialogs import MeshEquilibrationDialog
        from ..meshing import equilibrate_fit

        geometries = self.cdata.models.get_selected_geometries()
        if len(geometries) != 1:
            msg = "Can only equilibrate a single mesh at a time."
            return QMessageBox.warning(self, "Error", msg)

        geometry = geometries[0]
        if not hasattr(geometry.model, "mesh"):
            msg = f"{geometry} is not a triangular mesh."
            return QMessageBox.warning(self, "Error", msg)

        dialog = _getExistingDirectory(self, "Select or create directory")
        if dialog.exec_() != QFileDialog.Accepted:
            return None

        directory = dialog.selectedFiles()[0]
        if not directory:
            return -1

        dialog = MeshEquilibrationDialog(None)
        if not dialog.exec():
            return -1

        submit_task(
            "Equilibrate",
            equilibrate_fit,
            None,
            geometry,
            directory,
            dialog.get_parameters(),
        )

    def _setup_hmff(self):
        from ..meshing import setup_hmff
        from ..dialogs import HMFFDialog

        dialog = _getExistingDirectory(
            self, "Select directory with equilibrated meshes."
        )
        if dialog.exec_() != QFileDialog.Accepted:
            return None

        directory = dialog.selectedFiles()[0]
        if not directory:
            return -1

        mesh_config = join(directory, "mesh.txt")
        if not exists(mesh_config):
            msg = f"Missing mesh_config at {mesh_config}. Most likely {directory} "
            "is not a valid directory created by Equilibrate Mesh."
            return QMessageBox.warning(self, "Error", msg)

        with open(mesh_config, mode="r", encoding="utf-8") as infile:
            data = [x.strip() for x in infile.read().split("\n")]
            data = [x.split("\t") for x in data if len(x)]

        headers = data.pop(0)
        ret = {header: list(column) for header, column in zip(headers, zip(*data))}

        if not all(t in ret.keys() for t in ("file", "scale_factor", "offset")):
            print(
                "mesh_config is malformated. Expected file, scale_factor, "
                f"offset columns, got {', '.join(list(ret.keys()))}."
            )
            return -1

        dialog = HMFFDialog(None, mesh_options=ret["file"])
        if not dialog.exec():
            return -1

        submit_task(
            "HMFF Setup",
            setup_hmff,
            None,
            ret,
            directory=directory,
            **dialog.get_parameters(),
        )

    def _import_trajectory(
        self,
        scale: float = 1.0,
        offset: Union[str, float] = 0.0,
        drop_pbc: bool = False,
        **kwargs,
    ):
        from ..meshing import to_open3d
        from ..formats import open_file
        from ..dialogs import ProgressDialog
        from ..geometry import GeometryTrajectory
        from ..parametrization import TriangularMesh

        dialog = _getExistingDirectory(self, "Select directory with DTS trajectory")
        if dialog.exec_() != QFileDialog.Accepted:
            return None

        directory = dialog.selectedFiles()[0]
        if not directory:
            return -1

        ret = []
        files = [join(directory, x) for x in listdir(directory)]
        files = [
            x
            for x in files
            if x.endswith(".tsi") or x.endswith(".vtu") and x != "conf-1.vtu"
        ]
        files = sorted(files, key=lambda x: int(re.findall(r"\d+", basename(x))[0]))

        if isinstance(offset, str):
            try:
                offset = np.array([float(x) for x in offset.split(",")])
            except Exception as e:
                raise ValueError(
                    "Offset should be a single or three comma-separated floats."
                ) from e

        with ProgressDialog(files, title="Importing Trajectory", parent=None) as pbar:
            for index, filename in enumerate(pbar):
                container = open_file(filename)[0]
                faces = container.faces.astype(int)
                points = np.divide(np.subtract(container.vertices, offset), scale)

                if drop_pbc:
                    from ..meshing.utils import _edge_lengths

                    points_norm = points - points.min(axis=0)

                    box_stop = points_norm.max(axis=0)
                    points_pbc = np.mod(points_norm, 0.85 * box_stop)

                    dist_regular = _edge_lengths(points_norm, faces)
                    dist_pbc = _edge_lengths(points_pbc, faces)

                    keep = np.all(dist_pbc >= dist_regular, axis=-1)
                    faces = faces[keep]

                # Avoid detecting PBC as ill-defined meshes
                fit = TriangularMesh(to_open3d(points, faces), repair=False)

                ret.append(
                    {
                        "fit": fit,
                        "filename": filename,
                        "name": basename(directory),
                        "vertex_properties": container.vertex_properties,
                    }
                )

        if len(ret) == 0:
            raise ValueError(f"No meshes found at: {directory}.")

        base = ret[0]["fit"]
        trajectory = GeometryTrajectory(
            points=base.vertices.copy(),
            normals=base.compute_vertex_normals().copy(),
            sampling_rate=1 / scale,
            meta=ret[0].copy(),
            trajectory=ret,
            vertex_properties=ret[0]["vertex_properties"],
            model=base,
        )
        trajectory.change_representation("mesh")
        self.cdata._models.add(trajectory)
        self.cdata.models.data_changed.emit()
        return self.cdata.models.render()

    def _map_fit(self):
        from ..meshing import mesh_to_cg
        from ..dialogs import MeshMappingDialog

        dialog = _getExistingDirectory(self, "Select output directory")
        if dialog.exec_() != QFileDialog.Accepted:
            return None

        directory = dialog.selectedFiles()[0]
        if not directory:
            return -1

        fits = self.cdata.format_datalist("models", mesh_only=True)
        clusters = self.cdata.format_datalist("data")
        dialog = MeshMappingDialog(fits=fits, clusters=clusters)
        if not dialog.exec():
            return -1

        fit, edge_length, mappings, cast_ray, flip = dialog.get_parameters()

        submit_task(
            "Coarse graining",
            mesh_to_cg,
            None,
            edge_length=edge_length,
            output_directory=directory,
            inclusions=mappings,
            include_normals=cast_ray,
            flip_normals=flip,
        )

    def _run_membrain(self, *args, **kwargs):
        from ..gui import App
        from ..formats import open_file
        from ..segmentation import run_membrainseg

        def _callback(output_name: str):
            if output_name is None:
                return QMessageBox.warning(
                    None, "Error", "No segmentation was created."
                )

            # Preferred because it also updates viewport
            app = QApplication.instance().activeWindow()
            if isinstance(app, App):
                return app._open_files([output_name])

            container = open_file(output_name)
            for data in container:
                self.cdata.data.add(
                    points=data.vertices,
                    normals=data.normals,
                    sampling_rate=data.sampling,
                )
            self.cdata.data.data_changed.emit()
            self.cdata.data.render()

        submit_task(
            "Membrane Segmentation", run_membrainseg, _callback, *args, **kwargs
        )

    def _run_membrane_segmentation(self, **kwargs):
        file_name, _ = QFileDialog.getOpenFileName(
            self, "Select volume", "", "MRC Files (*.mrc);;All Files (*.*)"
        )
        if not file_name:
            return None

        if not exists(kwargs.get("model_path", "")):
            return QMessageBox.warning(None, "Error", "Missing path to membrain model.")

        return self._run_membrain(tomogram_path=file_name, **kwargs)

    def add_cloud(self, *args):
        num_points = 1000
        points = np.random.rand(num_points, 3) * 100
        self.cdata.data.add(points=points, sampling_rate=1)

        self.cdata.data.render()
        # self.cdata.models.render()


IMPORT_SETTINGS = {
    "title": "Trajectory Import",
    "settings": [
        {
            "label": "Scale",
            "parameter": "scale",
            "type": "text",
            "default": 1.0,
            "description": "Scale imported points by 1 / scale.",
        },
        {
            "label": "Offset",
            "parameter": "offset",
            "type": "text",
            "default": "0.0",
            "description": "Add offset as (points - offset) / scale ",
        },
        {
            "label": "Remove PBC",
            "parameter": "drop_pbc",
            "type": "boolean",
            "default": False,
            "description": "Drop triangles arising from periodic boundaries.",
        },
    ],
}



"""
Parallel backend for offloading compute heavy tasks.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import uuid
import warnings
import concurrent
from typing import Callable, Any, Dict

from qtpy.QtWidgets import QMessageBox
from qtpy.QtCore import QObject, Signal, QTimer, QThread


def _default_messagebox(task_name: str, msg: str, is_warning: bool = False):
    readable_name = task_name.replace("_", " ").title()

    msg_box = QMessageBox()
    msg_box.setIcon(QMessageBox.Icon.Critical)
    msg_box.setWindowTitle("Operation Failed")
    msg_box.setText(f"{readable_name} Failed with Errors")
    if is_warning:
        msg_box.setIcon(QMessageBox.Icon.Warning)
        msg_box.setWindowTitle("Operation Warning")
        msg_box.setText(f"{readable_name} Completed with Warnings")

    msg_box.setInformativeText(str(msg))
    msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)
    msg_box.exec()


def _wrap_warnings(func, *args, **kwargs):
    """Wrapper function that captures warnings and returns them with the result"""
    with warnings.catch_warnings(record=True) as warning_list:
        warnings.simplefilter("always")

        try:
            result = func(*args, **kwargs)

            warning_msg = ""
            for warning_item in warning_list:

                # TODO: Manage citation warnings more rigorously
                if "citation" in str(warning_item.message).lower():
                    continue

                if warning_item.category is DeprecationWarning:
                    continue

                warning_msg += (
                    f"{warning_item.category.__name__}: {warning_item.message}\n"
                )

            return {
                "result": result,
                "warnings": warning_msg.rstrip() if warning_msg else None,
            }

        except Exception as e:
            # Re-raise the exception so it's handled by the executor
            raise e


class BatchContext:
    def __init__(self, render_callback, delay: 500):
        self.manager = BackgroundTaskManager.instance()
        self.render_callback = render_callback
        self.task_ids = []

        self._delay = 500

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        QTimer.singleShot(self._delay, self._check_completion)

    def _check_completion(self):
        # Check if all tasks done
        all_done = all(task_id not in self.manager.futures for task_id in self.task_ids)

        if all_done:
            self.render_callback()
        else:
            QTimer.singleShot(self._delay, self._check_completion)

    def submit_task(self, name, func, callback, *args, **kwargs):
        task_id = self.manager.submit_task(name, func, callback, *args, **kwargs)
        self.task_ids.append(task_id)
        return task_id


class BackgroundTaskManager(QObject):
    task_started = Signal(str, str)  # task_id, task_name
    task_completed = Signal(str, str, object)  # task_id, task_name, result
    task_failed = Signal(str, str, str)  # task_id, task_name, error
    task_warning = Signal(str, str, str)  # task_id, task_name, warning

    running_tasks = Signal(int)  # running tasks

    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = BackgroundTaskManager()
        return cls._instance

    def __init__(self):
        super().__init__()

        self.executor = concurrent.futures.ProcessPoolExecutor(
            max_workers=max(2, QThread.idealThreadCount() - 1)
        )

        self.task_info: Dict[str, Dict[str, Any]] = {}
        self.futures: Dict[str, concurrent.futures.Future] = {}

        self.timer = QTimer()
        self.timer.timeout.connect(self._check_completed_tasks)
        self.timer.start(300)

        self.task_failed.connect(self._default_error_handler)
        self.task_warning.connect(self._default_warning_handler)

    def _default_error_handler(self, task_id, task_name, error):
        """Default handler for task errors."""
        return _default_messagebox(task_name, error, is_warning=False)

    def _default_warning_handler(self, task_id, task_name, warning):
        """Default handler for task errors."""
        return _default_messagebox(task_name, warning, is_warning=True)

    def submit_task(
        self, name: str, func: Callable, callback: Callable = None, *args, **kwargs
    ) -> str:
        """Submit a task to the executor"""
        task_id = str(uuid.uuid4())

        self.task_info[task_id] = {"name": name, "callback": callback}

        self.futures[task_id] = self.executor.submit(
            _wrap_warnings, func, *args, **kwargs
        )

        self.task_started.emit(task_id, name)
        self.running_tasks.emit(len(self.futures))
        return task_id

    def _check_completed_tasks(self):
        """Check for completed futures and handle results"""
        completed_tasks = []

        for task_id, future in self.futures.items():
            if future.done():
                task_info = self.task_info[task_id]
                task_name = task_info["name"]

                try:
                    ret = future.result()

                    result = ret["result"]
                    warnings_msg = ret["warnings"]

                    self.task_completed.emit(task_id, task_name, result)

                    if task_info["callback"]:
                        task_info["callback"](result)

                    if warnings_msg is not None:
                        self.task_warning.emit(task_id, task_name, warnings_msg)

                except Exception as e:
                    error_msg = str(e)
                    self.task_failed.emit(task_id, task_name, error_msg)
                completed_tasks.append(task_id)

        for task_id in completed_tasks:
            _ = self.futures.pop(task_id)
            _ = self.task_info.pop(task_id)
            self.running_tasks.emit(len(self.futures))


def submit_task(name, func, callback, *args, **kwargs):
    return BackgroundTaskManager.instance().submit_task(
        name, func, callback, *args, **kwargs
    )



from typing import Dict

import numpy as np
from tme import Orientations, Density
from scipy.spatial.transform import Rotation

from ._utils import get_extension


class OrientationsWriter:
    def __init__(
        self, points: np.ndarray, quaternions: np.ndarray, entities: np.ndarray
    ):
        """
        Initialize writer with point coordinates, quaternions, and entity labels.

        Parameters
        ----------
        points : np.ndarray
            Array of 3D point coordinates.
        quaternions : np.ndarray
            Array of quaternion rotations.
        entities : np.ndarray
            Array of entity labels for each point.
        """
        self.entities = entities
        self.points = points
        rotations = Rotation.from_quat(quaternions, scalar_first=True).inv()
        self.rotations = rotations.as_euler(seq="ZYZ", degrees=True)

    def to_file(self, file_path, file_format: str = None, **kwargs):
        """
        Write orientations data to file in specified format.

        Parameters
        ----------
        file_path : str
            Output file path.
        file_format : str, optional
            Output format, inferred from extension if None.
        **kwargs
            Additional keyword arguments passed to writer.

        Raises
        ------
        ValueError
            If the file format is not supported.
        """
        _supported_formats = ("tsv", "star")

        if file_format is None:
            file_format = get_extension(file_path)[1:]

        if file_format not in _supported_formats:
            formats = ", ".join([str(x) for x in _supported_formats])
            raise ValueError(f"Supported formats are {formats}.")
        return self._write_orientations(file_path, **kwargs)

    def _write_orientations(self, file_path, **kwargs):
        """
        Backend function for writing orientations to file.

        Parameters
        ----------
        file_path : str
            Output file path.
        **kwargs
            Additional keyword arguments passed to orientations writer.
        """
        orientations = Orientations(
            translations=self.points,
            rotations=self.rotations,
            scores=np.zeros(self.rotations.shape[0]),
            details=self.entities,
        )
        return orientations.to_file(file_path, **kwargs)


def write_density(
    data: np.ndarray, filename: str, sampling_rate: float = 1, origin: float = 0
) -> None:
    """
    Write 3D density data to file (typically in CCP4/MRC format).

    Parameters
    ----------
    data : np.ndarray
        3D density array.
    filename : str
        Output file path.
    sampling_rate : float, optional
        Sampling rate per voxel, by default 1 Angstrom / Voxel.
    origin : float, optional
        Origin offset for the density data in Angstrom, by default 0.
    """
    return Density(data, sampling_rate=sampling_rate, origin=origin).to_file(filename)


def write_topology_file(file_path: str, data: Dict, tsi_format: bool = False) -> None:
    """
    Write a topology file [1]_.

    Parameters
    ----------
    file_path : str
        The path to the output file.
    data : dict
        Topology file data as per :py:meth:`read_topology_file`.
    tsi_format : bool, optional
        Whether to use the '.q' or '.tsi' file, defaults to '.q'.

    References
    ----------
    .. [1] https://github.com/weria-pezeshkian/FreeDTS/wiki/Manual-for-version-1
    """
    vertex_string = f"{data['vertices'].shape[0]}\n"
    if tsi_format:
        vertex_string = f"vertex {vertex_string}"

    stop = data["vertices"].shape[1] - 1
    if tsi_format:
        stop = data["vertices"].shape[1]
    for i in range(data["vertices"].shape[0]):
        vertex_string += f"{int(data['vertices'][i, 0])}  "
        vertex_string += "  ".join([f"{x:<.10f}" for x in data["vertices"][i, 1:stop]])

        if not tsi_format:
            vertex_string += f"  {int(data['vertices'][i, stop])}"
        vertex_string += "\n"

    stop = data["faces"].shape[1] - 1
    face_string = f"{data['faces'].shape[0]}\n"
    if tsi_format:
        face_string = f"triangle {face_string}"
        stop = data["faces"].shape[1]
    for i in range(data["faces"].shape[0]):
        face = [f"{int(x):d}" for x in data["faces"][i, :stop]]
        face[1] += " "
        face.append("")
        face_string += "  ".join(face) + "\n"

    inclusion_string = ""
    inclusions = data.get("inclusions", None)
    if tsi_format and inclusions is not None:
        inclusion_string = f"inclusion {inclusions.shape[0]}\n"
        for i in range(data["inclusions"].shape[0]):
            ret = inclusions[i]
            ret[0] = int(ret[0])
            ret[1] = int(ret[1])
            ret[2] = int(ret[2])
            inclusion_string += f"{'   '.join([f'{x}' for x in ret])}   \n"

    box_string = f"{'   '.join([f'{x:<.10f}' for x in data['box']])}   \n"
    if tsi_format:
        box_string = f"version 1.1\nbox   {box_string}"

    with open(file_path, mode="w", encoding="utf-8") as ofile:
        ofile.write(box_string)
        ofile.write(vertex_string)
        ofile.write(face_string)
        ofile.write(inclusion_string)



from .hmff import HMFFDialog
from .export import ExportDialog
from .keybinds import KeybindsDialog
from .progress import ProgressDialog
from .histogram import HistogramDialog
from .cropping import DistanceCropDialog
from .import_data import ImportDataDialog
from .appsettings import AppSettingsDialog
from .backmapping import MeshMappingDialog
from .tilt_control import TiltControlDialog
from .matching import TemplateMatchingDialog
from .update import UpdateChecker, UpdateDialog
from .properties import GeometryPropertiesDialog
from .equilibration import MeshEquilibrationDialog
from .property_analysis import PropertyAnalysisDialog
from .operation import *



from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSpinBox,
    QCheckBox,
    QComboBox,
    QGroupBox,
    QFormLayout,
    QDoubleSpinBox,
    QGridLayout,
    QPushButton,
)

from mosaic.widgets import create_setting_widget


class AnimationSettings(QGroupBox):
    animationChanged = Signal(dict)

    def __init__(self, parent=None):
        super().__init__("Properties", parent)
        self.animation = None
        self.parameter_widgets = {}
        self.setup_ui()

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(10, 10, 10, 10)

        name_layout = QHBoxLayout()
        name_layout.addWidget(QLabel("Name:"))
        self.name_edit = QLineEdit()
        self.name_edit.setMaximumWidth(140)
        self.name_edit.textChanged.connect(lambda x: self.on_change(x, "name"))
        name_layout.addWidget(self.name_edit)

        name_layout.addWidget(QLabel("Enabled:"))
        self.enabled_check = QCheckBox()
        self.enabled_check.setChecked(True)
        self.enabled_check.stateChanged.connect(
            lambda x: self.on_change(x, key="enabled")
        )
        name_layout.addWidget(self.enabled_check)
        main_layout.addLayout(name_layout)

        frame_group = QGroupBox("Frames")
        frame_layout = QGridLayout(frame_group)

        frame_layout.addWidget(QLabel("Global Start:"), 0, 0)
        self.global_start_spin = QSpinBox()
        self.global_start_spin.setRange(0, 2 << 29)
        self.global_start_spin.valueChanged.connect(
            lambda x: self.on_change(x, "global_start_frame")
        )
        frame_layout.addWidget(self.global_start_spin, 0, 1)

        frame_layout.addWidget(QLabel("Stride:"), 1, 0)
        self.stride_spin = QSpinBox()
        self.stride_spin.setRange(1, 2 << 29)
        self.stride_spin.setValue(1)
        self.stride_spin.valueChanged.connect(lambda x: self.on_change(x, "stride"))
        frame_layout.addWidget(self.stride_spin, 1, 1)

        frame_layout.addWidget(QLabel("Local Start:"), 2, 0)
        self.start_spin = QSpinBox()
        self.start_spin.setRange(0, 10000)
        self.start_spin.valueChanged.connect(lambda x: self.on_change(x, "start_frame"))
        frame_layout.addWidget(self.start_spin, 2, 1)

        frame_layout.addWidget(QLabel("Local Stop:"), 3, 0)
        self.stop_spin = QSpinBox()
        self.stop_spin.setRange(0, 10000)
        self.stop_spin.valueChanged.connect(lambda x: self.on_change(x, "stop_frame"))
        frame_layout.addWidget(self.stop_spin, 3, 1)

        main_layout.addWidget(frame_group)

        self.params_group = QGroupBox("Parameters")
        self.params_layout = QFormLayout(self.params_group)
        self.params_layout.setContentsMargins(10, 10, 10, 10)
        main_layout.addWidget(self.params_group)

        main_layout.addStretch()

    def set_animation(self, animation):
        self.animation = animation
        self.name_edit.setText(animation.name)
        self.global_start_spin.setValue(animation.global_start_frame)
        self.start_spin.setValue(animation.start_frame)
        self.stop_spin.setValue(animation.stop_frame)
        self.stop_spin.setMaximum(animation.frames)
        self.stride_spin.setValue(animation.stride)
        self.enabled_check.setChecked(animation.enabled)

        while self.params_layout.rowCount() > 0:
            self.params_layout.removeRow(0)
        self.parameter_widgets.clear()

        for widget_settings in animation.get_settings():
            if widget_settings["type"] == "button":
                widget = QPushButton(widget_settings["text"])
                widget.clicked.connect(widget_settings["callback"])
            else:
                widget = create_setting_widget(widget_settings)

            signal = None
            if isinstance(widget, QComboBox):
                signal = widget.currentTextChanged
            elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
                signal = widget.valueChanged
            elif isinstance(widget, QLineEdit):
                signal = widget.textChanged

            label = widget_settings["label"]

            if signal is not None:
                signal.connect(lambda x, lab=label: self.on_change(x, lab))

            label_clean = label.title().replace("_", " ")
            self.params_layout.addRow(f"{label_clean}:", widget)
            self.parameter_widgets[label] = widget

    def on_change(self, value, key):
        if not self.animation:
            return None

        attr = getattr(self.animation, key, None)
        if attr is not None:
            setattr(self.animation, key, value)
        else:
            self.animation.update_parameters(**{key: value})

        if self.animation.frames != self.stop_spin.maximum():
            self.start_spin.setMaximum(self.animation.frames)
            self.stop_spin.setMaximum(self.animation.frames)

            if self.animation.frames < (2 << 15):
                self.stop_spin.setValue(self.animation.frames)

        self.animationChanged.emit({key: value})


class ExportSettings(QGroupBox):
    def __init__(self, parent=None):
        super().__init__("Export Settings", parent)
        self.setup_ui()

    def setup_ui(self):
        layout = QGridLayout(self)

        layout.addWidget(QLabel("Format:"), 0, 0)
        self.format_combo = QComboBox()
        self.format_combo.addItems(["MP4", "AVI", "RGBA"])
        layout.addWidget(self.format_combo, 0, 1)

        layout.addWidget(QLabel("Quality:"), 1, 0)
        self.quality_spin = QSpinBox()
        self.quality_spin.setRange(0, 100)
        self.quality_spin.setValue(80)
        self.quality_spin.setSuffix("%")
        layout.addWidget(self.quality_spin, 1, 1)

        layout.addWidget(QLabel("Rate (fps):"), 2, 0)
        self.frame_rate = QSpinBox()
        self.frame_rate.setRange(1, 1000)
        self.frame_rate.setValue(30)
        layout.addWidget(self.frame_rate, 2, 1)

        layout.addWidget(QLabel("Stride:"), 3, 0)
        self.frame_stride = QSpinBox()
        self.frame_stride.setRange(1, 100)
        self.frame_stride.setValue(1)
        layout.addWidget(self.frame_stride, 3, 1)

        layout.addWidget(QLabel("Window:"), 4, 0)
        range_layout = QHBoxLayout()
        self.start_frame = QSpinBox()
        self.start_frame.setFixedWidth(70)
        self.start_frame.setRange(0, 10000)
        self.end_frame = QSpinBox()
        self.end_frame.setFixedWidth(70)
        self.end_frame.setRange(0, 10000)
        self.end_frame.setValue(300)
        range_layout.addWidget(self.start_frame)
        range_layout.addWidget(self.end_frame)
        layout.addLayout(range_layout, 4, 1)



from typing import Optional

from qtpy.QtCore import Qt, Signal, QPropertyAnimation, QEasingCurve
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QToolButton,
    QLabel,
    QFrame,
    QSizePolicy,
    QScrollArea,
    QSplitter,
)
import qtawesome as qta


class ChevronButton(QToolButton):
    def __init__(self, direction="right", parent=None):
        """Initialize a chevron button with specified direction."""
        super().__init__(parent)
        self.direction = direction
        self.setFixedSize(24, 24)
        self.setStyleSheet(
            """
            QToolButton {
                border: none;
                background-color: transparent;
            }
            QToolButton:hover {
                background-color: rgba(0, 0, 0, 0.05);
                border-radius: 12px;
            }
        """
        )
        self._update_icon()

    def set_direction(self, direction):
        """Change the direction of the chevron."""
        self.direction = direction
        self._update_icon()

    def _update_icon(self):
        self.setIcon(qta.icon("fa5s.chevron-right", color="#696c6f"))
        if self.direction == "left":
            self.setIcon(qta.icon("fa5s.chevron-left", color="#696c6f"))
        elif self.direction == "up":
            self.setIcon(qta.icon("fa5s.chevron-up", color="#696c6f"))
        elif self.direction == "down":
            self.setIcon(qta.icon("fa5s.chevron-down", color="#696c6f"))
        return None


class ObjectBrowserSidebarSection(QWidget):
    """A simple section with a header that can contain any widget."""

    def __init__(self, title: str, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.title = title

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        header_frame = QFrame()
        header_frame.setObjectName("sectionHeader")
        header_frame.setFixedHeight(26)

        header_layout = QHBoxLayout(header_frame)
        header_layout.setContentsMargins(8, 0, 8, 0)

        title_label = QLabel(title)
        title_label.setObjectName("sectionTitle")

        header_layout.addWidget(title_label)
        header_layout.addStretch()

        self.content_widget = QWidget()
        self.content_widget.setObjectName("sectionContent")

        self.content_layout = QVBoxLayout(self.content_widget)
        self.content_layout.setContentsMargins(0, 0, 0, 0)
        self.content_layout.setSpacing(0)

        layout.addWidget(header_frame)
        layout.addWidget(self.content_widget, 1)

    def addWidget(self, widget):
        """Add a widget to the content layout."""
        self.content_layout.addWidget(widget)


class ObjectBrowserSidebar(QWidget):
    """
    Simplified sidebar component for Mosaic.
    Provides direct widget support for maximum flexibility.
    """

    visibility_changed = Signal(bool)

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.visible = True
        self.collapsed = False
        self.sections = {}

        self._setup_ui()
        self._setup_styling()

        self.animation = QPropertyAnimation(self, b"maximumWidth")
        self.animation.setDuration(200)
        self.animation.setEasingCurve(QEasingCurve.Type.InOutQuad)
        self.animation.valueChanged.connect(self._update_width)
        self.animation.finished.connect(self._animation_finished)

    def _update_width(self, width):
        """Update the sidebar width during animation."""
        show_content = width > 50
        if hasattr(self, "title_label"):
            self.title_label.setVisible(show_content)

        if hasattr(self, "scroll_area"):
            self.scroll_area.setVisible(show_content)

    def _animation_finished(self):
        """Handle animation completion."""
        if not self.collapsed:
            # Avoid unwanted splitter interactions
            self.setMaximumWidth(16777215)

    def _setup_ui(self):
        """Initialize the UI components."""
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        # Header section
        self.header = QWidget()
        self.header.setObjectName("sidebarHeader")
        self.header.setFixedHeight(40)
        header_layout = QHBoxLayout(self.header)
        header_layout.setContentsMargins(10, 8, 10, 8)

        self.title_label = QLabel()
        self.title_label.setObjectName("panelTitle")
        self.collapse_btn = ChevronButton("left")
        self.collapse_btn.clicked.connect(self.toggle_collapse)

        header_layout.addWidget(self.title_label)
        header_layout.addStretch()
        header_layout.addWidget(self.collapse_btn)
        main_layout.addWidget(self.header)

        # Content scroll area
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFrameShape(QFrame.Shape.NoFrame)
        self.scroll_area.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.scroll_area.setObjectName("scrollArea")

        # Content widget
        self.content_widget = QWidget()
        self.content_widget.setMinimumWidth(100)
        self.content_layout = QVBoxLayout(self.content_widget)
        self.content_layout.setContentsMargins(6, 6, 6, 6)
        self.content_layout.setSpacing(10)

        self.scroll_area.setWidget(self.content_widget)
        main_layout.addWidget(self.scroll_area, 1)
        main_layout.setStretchFactor(self.scroll_area, 1)
        main_layout.addStretch(0)

        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)

    def _setup_styling(self):
        """Set up the widget styling."""
        self.setStyleSheet(
            """
            QLabel {
                background-color: transparent;
            }

            /* Header styling with border-bottom */
            #sidebarHeader {
                background-color: transparent;
                border-bottom: 1px solid #6b7280;
            }

            #panelTitle {
                font-weight: 600;
                font-size: 13px;
            }

            /* Section styling */
            #sectionHeader {
                background-color: #1a000000;
                border-radius: 4px;
            }

            #sectionContent {
                background-color: transparent;
            }

            #sectionTitle {
                font-weight: 500;
                font-size: 13px;
            }

            /* Item styling */
            #selectedItem {
                background-color: #eef2ff;
                border-radius: 4px;
            }

            #normalItem {
                background-color: transparent;
                border-radius: 4px;
            }

            #normalItem:hover {
                background-color: #f5f5f5;
            }

            #visibilityDot[status="visible"] {
                background-color: #34d399;
                border-radius: 5px;
            }

            #visibilityDot[status="hidden"] {
                background-color: #9ca3af;
                border-radius: 5px;
            }

            #metadataLabel {
                color: #6b7280;
                font-size: 10px;
            }

            /* Content area */
            #scrollArea {
                background-color: transparent;
                border: none;
            }

            #contentWidget {
                background-color: transparent;
            }

            QScrollArea {
                border: none;
            }

            QScrollBar:vertical {
                border: none;
                background: #f1f1f1;
                width: 6px;
                margin: 0px;
            }

            QScrollBar::handle:vertical {
                background: #c1c1c1;
                min-height: 20px;
                border-radius: 3px;
            }

            QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {
                height: 0px;
            }

            QScrollBar:horizontal {
                height: 0px;
                background: transparent;
            }
        """
        )

    def add_widget(self, section_id: str, title: str, widget: QWidget) -> QWidget:
        """Add a widget wrapped in a section with header to the sidebar."""
        section = ObjectBrowserSidebarSection(title)
        widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        section.addWidget(widget)

        for i in range(self.content_layout.count()):
            item = self.content_layout.itemAt(i)
            if item and item.spacerItem():
                self.content_layout.removeItem(item)

        self.content_layout.addWidget(section, 1)
        self.content_layout.addStretch(0)

        self.sections[section_id] = section
        return widget

    def clear_sections(self):
        """Remove all sections."""
        for section_id in list(self.sections.keys()):
            self.remove_section(section_id)

    def toggle_collapse(self):
        """Toggle between collapsed and expanded states with animation."""
        splitter = self.parent()

        while splitter and not isinstance(splitter, QSplitter):
            splitter = splitter.parent()

        if not splitter:
            return None

        current_sizes = splitter.sizes()
        if not self.collapsed:
            # Store current width before collapsing
            self.previous_width = current_sizes[0]
            target_sizes = [60, current_sizes[1] + current_sizes[0] - 60]
            direction = "right"
        else:
            # Restore previous width
            target_width = getattr(self, "previous_width", 200)
            target_sizes = [
                target_width,
                current_sizes[1] - target_width + current_sizes[0],
            ]
            direction = "left"

        # Animate splitter sizes instead of widget width
        self.animation.setStartValue(current_sizes[0])
        self.animation.setEndValue(target_sizes[0])
        self.animation.valueChanged.connect(
            lambda width: self._animate_splitter_size(width, splitter)
        )

        self.collapsed = not self.collapsed
        self.animation.start()
        self.collapse_btn.set_direction(direction)
        self.visibility_changed.emit(self.collapsed)

    def _animate_splitter_size(self, width, splitter):
        """Update splitter size during animation."""
        current_sizes = splitter.sizes()
        total_width = sum(current_sizes)
        new_sizes = [int(width), total_width - int(width)]
        splitter.setSizes(new_sizes)

        # Show/hide content based on width
        show_content = width > 60
        if hasattr(self, "title_label"):
            self.title_label.setVisible(show_content)
        if hasattr(self, "scroll_area"):
            self.scroll_area.setVisible(show_content)

    def set_title(self, title: str):
        """Set the sidebar title."""
        self.title_label.setText(title)



#!python3
"""
GUI entrypoint.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import sys
import argparse
from importlib_resources import files

from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication

from mosaic import __version__
from mosaic.stylesheets import (
    QMessageBox_style,
    QLineEdit_style,
    QSpinBox_style,
    QDoubleSpinBox_style,
    QComboBox_style,
    QCheckBox_style,
    QSlider_style,
    QGroupBox_style,
    QListWidget_style,
    QToolButton_style,
    QMenu_style,
)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=f"{__version__}")
    parser.parse_args()

    app = QApplication(sys.argv)
    app.setApplicationName("Mosaic")
    app.setApplicationDisplayName("Mosaic")

    icon = QIcon(str(files("mosaic.data").joinpath("data/mosaic.icns")))
    app.setWindowIcon(icon)

    # Fixes alignment issue in default style
    # https://forum.qt.io/topic/105191/why-isn-t-a-qcombobox-positioned-correctly-in-a-layout/11
    app.setStyle("Fusion")
    app.setStyleSheet(
        QMessageBox_style
        + QLineEdit_style
        + QSpinBox_style
        + QDoubleSpinBox_style
        + QComboBox_style
        + QCheckBox_style
        + QSlider_style
        + QGroupBox_style
        + QListWidget_style
        + QToolButton_style
        + QMenu_style
    )

    from mosaic.gui import App

    window = App()
    window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



"""
Implements DataContainer as handler of Geometry object collections.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import List, Tuple, Union, Dict

import numpy as np

__all__ = ["DataContainer"]


class DataContainer:
    """
    Container for managing and manipulating point cloud data collections.

    Parameters
    ----------
    base_color : tuple of float, optional
        Default color for points in RGB format in range 0-1.
        Default is (0.7, 0.7, 0.7).
    highlight_color : tuple of float, optional
        Highlight color for points in RGB format in range 0-1.
        Default is (0.8, 0.2, 0.2).
    """

    def __init__(self, base_color=(0.7, 0.7, 0.7), highlight_color=(0.8, 0.2, 0.2)):
        self.data = []
        self.metadata = {}
        self.base_color = base_color
        self.highlight_color = highlight_color

    def __len__(self):
        return len(self.data)

    def get_actors(self):
        """Get VTK actors from all geometries.

        Returns
        -------
        list
            List of VTK actors.
        """
        return [x.actor for x in self.data]

    def add(self, points, color=None, **kwargs):
        """Add a new geometry object to the container.

        Parameters
        ----------
        points : np.ndarray or Geometry
            Points to add to the container.
        color : tuple of float, optional
            RGB color values for the point cloud.

        Returns
        -------
        int
            Index of the new point cloud.
        """
        from .geometry import Geometry

        if color is None:
            color = self.base_color

        if issubclass(type(points), Geometry):
            new_geometry = points
        else:
            new_geometry = Geometry(points, color=color, **kwargs)

        new_geometry.set_appearance(
            base_color=color, highlight_color=self.highlight_color
        )
        self.data.append(new_geometry)
        return len(self.data) - 1

    def remove(self, uuids_or_geometries: Union[List[str], List["Geometry"]]):
        """Remove geometries at specified indices or by geometry objects.

        Parameters
        ----------
        uuids_or_geometries : str, list of str, Geometry, or list of Geometry
            UUIDs of geometries to remove or geometry objects to remove.
        """
        indices = [self.uuid_to_index(x) for x in self._to_uuids(uuids_or_geometries)]
        indices = list(set(x for x in indices if self._index_ok(x)))

        # Reverse order to avoid potential shift issue
        for index in sorted(indices, reverse=True):
            self.data.pop(index)

    def clear(self):
        """Remove all data associated with the container."""
        self.data.clear()
        self.metadata.clear()

    def uuid_to_index(self, uuid: str) -> int:
        """Convert a uuid to an index in self.data."""
        for index, geometry in enumerate(self.data):
            if geometry.uuid == uuid:
                return index
        return None

    def get(self, index_or_uuid: Union[int, str]):
        """
        Retrieve the Geometry object by index or UUID.

        Parameters
        ----------
        index_or_uuid : int or str
            Geometry index or UUID to retrieve.

        Returns
        -------
        geometry : :py:class:`mosaic.geometry.Geometry`
            Selected geometry or None if not found.
        """
        if isinstance(index_or_uuid, str):
            index_or_uuid = self.uuid_to_index(index_or_uuid)

        if self._index_ok(index_or_uuid):
            return self.data[index_or_uuid]
        return None

    def update(self, uuid_or_geometry, new_geometry) -> bool:
        """
        Update a geometry by UUID or geometry object.

        Parameters
        ----------
        uuid_or_geometry : str or Geometry
            UUID of geometry to update, or the geometry object itself
        new_geometry : Geometry
            New geometry to replace the old one

        Returns
        -------
        bool
            True if update succeeded, False if UUID not found
        """
        uuid = self._to_uuid(uuid_or_geometry)
        if (index := self.uuid_to_index(uuid)) is None:
            return False

        if new_geometry.uuid != uuid:
            new_geometry.uuid = uuid

        self.data[index] = new_geometry
        return True

    def highlight_points(self, uuid_or_geometry, point_ids: set, color: Tuple[float]):
        """Highlight specific points in a cloud.

        Parameters
        ----------
        uuid_or_geometry : int or str
            UUID of geometry to update, or the geometry object itself
        point_ids : set
            IDs of points to highlight.
        color : tuple of float
            RGB color for highlighting.
        """
        if (geometry := self.get(self._to_uuid(uuid_or_geometry))) is None:
            return None

        if color is None:
            color = geometry._appearance.get("highlight_color", (0.8, 0.2, 0.2))
        geometry.color_points(point_ids, color)

    def highlight(self, uuids_or_geometries: Union[List[str], List["Geometry"]]):
        """Highlight specified geometries.

        Parameters
        ----------
        uuids_or_geometries : list of str or list of Geometry
            UUIDs or geometry objects to highlight.
        """
        uuids = self._to_uuids(uuids_or_geometries)
        _highlighted = getattr(self, "_highlighted_uuids", set())

        for geometry in self.data:
            appearance = geometry._appearance
            color = appearance.get("base_color", self.base_color)
            if geometry.uuid in uuids:
                color = appearance.get("highlight_color", self.highlight_color)
            elif geometry.uuid not in _highlighted:
                continue

            geometry.set_color(color=color)
        self._highlighted_uuids = set(uuids)

    def update_appearance(
        self, uuids_or_geometries: Union[List[str], List["Geometry"]], parameters: Dict
    ) -> bool:
        """Update appearance parameters for specified geometries.

        Parameters
        ----------
        uuids_or_geometries : list of str or list of Geometry
            UUIDs or geometry objects to update
        parameters : dict
            Appearance parameters to update

        Returns
        -------
        bool
            True if full render required (actor was replaced)
        """
        from .geometry import VolumeGeometry
        from .formats.parser import load_density

        volume = parameters.get("volume", None)
        volume_path = parameters.get("volume_path", None)
        if volume_path is not None:
            volume = load_density(volume_path)

        if volume is not None:
            sampling = volume.sampling_rate
            volume = volume.data * parameters.get("scale", 1.0)

        full_render = False
        parameters["isovalue_percentile"] = parameters.get("isovalue_percentile", 99.5)
        for uuid in self._to_uuids(uuids_or_geometries):
            if (geometry := self.get(uuid)) is None:
                continue

            if volume is not None:
                if not isinstance(geometry, VolumeGeometry):
                    geometry = geometry[...]
                state = geometry.__getstate__()

                try:
                    data_recent = np.allclose(state["volume"], volume)
                    # Check if representation has been switched in the meantime
                    data_recent = data_recent and geometry._representation == "volume"
                except Exception:
                    data_recent = False

                if not data_recent:
                    state["volume"] = volume
                    state["volume_sampling_rate"] = sampling

                    # New actor so make sure to re-render
                    full_render = True
                    geometry = VolumeGeometry(**state)
                    self.update(uuid, geometry)

            geometry.set_appearance(**parameters)

        return full_render

    def _to_uuid(self, uuid_or_geometry: Union[str, "Geometry"]) -> str:
        """Convert UUID or Geometry to UUID.

        Parameters
        ----------
        uuid_or_geometry : str or Geometry
            UUID or geometry object

        Returns
        -------
        str
            UUID string
        """
        from .geometry import Geometry

        if isinstance(uuid_or_geometry, Geometry):
            uuid_or_geometry = uuid_or_geometry.uuid
        return uuid_or_geometry

    def _to_uuids(
        self, uuids_or_geometries: Union[List[str], List["Geometry"]]
    ) -> List[str]:
        """Convert list of UUIDs or Geometries to UUIDs.

        Parameters
        ----------
        uuids_or_geometries : list of str or list of Geometry
            UUIDs or geometry objects

        Returns
        -------
        list of str
            UUIDs (strings)
        """
        if not isinstance(uuids_or_geometries, (list, tuple)):
            uuids_or_geometries = [uuids_or_geometries]
        return [self._to_uuid(x) for x in uuids_or_geometries]

    def _index_ok(self, index: int) -> bool:
        """Check if index is valid.

        Parameters
        ----------
        index : int
            Index to check.

        Returns
        -------
        bool
            True if index is valid.
        """
        try:
            index = int(index)
        except Exception:
            return False

        if 0 <= index < len(self.data):
            return True
        return False



import time

import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout

from ..widgets.ribbon import create_button


def noop(*args, **kwargs):
    return None


class PerformanceMonitor:
    """Monitor VTK rendering performance"""

    def __init__(self, render_window):
        self.render_window = render_window
        self.frame_times = []
        self.start_time = None

    def start_monitoring(self):
        """Start performance monitoring"""
        self.frame_times = []
        self.start_time = time.time()

    def record_frame(self):
        """Record a frame render time"""
        if self.start_time:
            frame_time = time.time() - self.start_time
            self.frame_times.append(frame_time)
            self.start_time = time.time()

    def get_stats(self):
        """Get performance statistics"""
        if not self.frame_times:
            return {}

        frame_times = np.array(self.frame_times)
        fps = 1.0 / np.mean(frame_times) if np.mean(frame_times) > 0 else 0

        return {
            "avg_fps": fps,
            "min_fps": 1.0 / np.max(frame_times) if np.max(frame_times) > 0 else 0,
            "max_fps": 1.0 / np.min(frame_times) if np.min(frame_times) > 0 else 0,
            "avg_frame_time_ms": np.mean(frame_times) * 1000,
            "frame_count": len(frame_times),
        }

    def print_stats(self, label="Performance"):
        """Print performance statistics"""
        stats = self.get_stats()
        if stats:
            print(f"\n=== {label} ===")
            print(f"Average FPS: {stats['avg_fps']:.1f}")
            print(f"Frame time: {stats['avg_frame_time_ms']:.1f}ms")
            print(f"FPS range: {stats['min_fps']:.1f} - {stats['max_fps']:.1f}")
            print(f"Frames measured: {stats['frame_count']}")


class DevelopmentTab(QWidget):
    def __init__(self, cdata, ribbon, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon
        self.legend = kwargs.get("legend", None)

        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

    def show_ribbon(self):
        self.ribbon.clear()
        cluster_actions = [
            create_button(
                "Add", "mdi.plus", self, self.add_cloud, "Merge selected clusters"
            ),
            create_button(
                "Test Render",
                "mdi.test-tube",
                self,
                self.test_point_rendering_performance,
                "Merge selected clusters",
            ),
        ]
        self.ribbon.add_section("Base Operations", cluster_actions)

    def add_cloud(self, *args):
        num_points = 50000
        points = np.random.rand(num_points, 3) * 100
        self.cdata._data.add(points=points, sampling_rate=1)
        self.cdata.data.data_changed.emit()
        self.cdata.data.render()

    def test_point_rendering_performance(self, *args, **kwargs):
        """
        Test rendering performance by spinning the camera
        """
        test_duration = 5.0
        vtk_widget = self.cdata.data.vtk_widget
        render_window = vtk_widget.GetRenderWindow()
        renderer = render_window.GetRenderers().GetFirstRenderer()
        camera = renderer.GetActiveCamera()

        monitor = PerformanceMonitor(render_window)
        monitor.start_monitoring()
        start_time = time.time()
        frame_count = 0

        while (time.time() - start_time) < test_duration:
            camera.Azimuth(1.0)

            monitor.start_time = time.time()
            render_window.Render()
            monitor.record_frame()

            frame_count += 1

            vtk_widget.GetRenderWindow().GetInteractor().ProcessEvents()

        return monitor



"""
Processing of Geometry objects.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import List, Optional
from functools import wraps

import numpy as np
from .utils import (
    statistical_outlier_removal,
    eigenvalue_outlier_removal,
    com_cluster_points,
    find_closest_points,
    connected_components,
    envelope_components,
    leiden_clustering,
    dbscan_clustering,
    birch_clustering,
    kmeans_clustering,
)

__all__ = ["GeometryOperations"]


def use_point_data(*, destructive=True):
    """
    Decorator to ensure operations work on underlying point cloud data.

    When a geometry is in mesh representation, operations should work on the
    original point cloud data (stored in _point_data), not the mesh vertices.
    This decorator transparently handles that conversion.
    """

    def decorator(operation):

        @wraps(operation)
        def wrapper(geometry, **kwargs):
            from .geometry import Geometry

            temp_geometry = geometry
            if is_mesh := geometry.is_mesh_representation():
                points, normals, quaternions = geometry.get_point_data()
                temp_geometry = Geometry(
                    points=points,
                    normals=normals,
                    quaternions=quaternions,
                    sampling_rate=geometry.sampling_rate,
                )

            # This is particularly useful for connected component clustering
            # on large segmentations to avoid recomputing meshes
            if is_mesh and not destructive:
                temp_geometry._cache["mesh"] = geometry._cache["mesh"]
                temp_geometry.change_representation(geometry._representation)

            return operation(temp_geometry, **kwargs)

        return wrapper

    return decorator


@use_point_data()
def decimate(geometry, method: str = "core", **kwargs):
    """
    Reduces the number of points in a point cloud by keeping only representative
    points based on the selected method.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str, optional
        Method to use. Options are:
        - 'outer' : Keep outer hull points using convex hull
        - 'core' : Keep core points using clustering
        - 'inner' : Keep inner points using spherical ray-casting
        Default is 'core'.
    **kwargs
        Additional arguments passed to the chosen method.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Decimated geometry.

    Raises
    ------
    ValueError
        If unsupported method is specified.
    """
    from .geometry import Geometry
    from .parametrization import ConvexHull

    points = geometry.points

    if method == "core":
        cutoff = kwargs.get("cutoff", None)
        if cutoff is None:
            cutoff = 4 * np.max(geometry._sampling_rate)

        points = com_cluster_points(points, cutoff)
    elif method == "outer":
        hull = ConvexHull.fit(
            points,
            elastic_weight=0,
            curvature_weight=0,
            volume_weight=0,
            voxel_size=geometry._sampling_rate,
        )
        hull_points = hull.sample(int(0.5 * points.shape[0]))
        _, indices = find_closest_points(points, hull_points)
        points = points[np.unique(indices)]
    elif method == "inner":
        # Budget ray-casting using spherical coordinates
        centroid = np.mean(points, axis=0)
        centered_points = points - centroid

        r = np.linalg.norm(centered_points, axis=1)
        theta = np.arccos(centered_points[:, 2] / r)
        phi = np.arctan2(centered_points[:, 1], centered_points[:, 0])

        n_phi_bins = 360
        theta_idx = np.digitize(theta, np.linspace(0, np.pi, n_phi_bins // 2))
        phi_idx = np.digitize(phi, np.linspace(-np.pi, np.pi, n_phi_bins))
        bin_id = theta_idx * n_phi_bins + phi_idx

        inner_indices = []
        for b in np.unique(bin_id):
            mask = np.where(bin_id == b)[0]
            inner_indices.append(mask[np.argmin(r[mask])])

        points = points[inner_indices]
    else:
        raise ValueError("Supported methods are 'inner', 'core' and 'outer'.")

    return Geometry(points, sampling_rate=geometry._sampling_rate)


@use_point_data()
def downsample(geometry, method: str = "radius", **kwargs):
    """
    Reduces point density by removing points based on spatial or random criteria.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str, optional
        Method to use. Options are:
        - 'radius' : Remove points that fall within radius of each other using voxel downsampling
        - 'number' : Randomly subsample points to target number
        Default is 'radius'.
    **kwargs
        Additional arguments passed to the chosen method:
        - For 'radius': voxel_size parameter for open3d.voxel_down_sample
        - For 'number': size parameter specifying target number of points

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Downsampled geometry.
    """
    from .geometry import Geometry

    points, normals = geometry.points, geometry.normals

    if method.lower() == "radius":
        import open3d as o3d

        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(points)
        pcd.normals = o3d.utility.Vector3dVector(normals)
        pcd = pcd.voxel_down_sample(**kwargs)
        points = np.asarray(pcd.points)
        normals = np.asarray(pcd.normals)
    elif method.lower() == "number":
        size = kwargs.get("size", 1000)
        size = min(size, points.shape[0])
        keep = np.random.choice(range(points.shape[0]), replace=False, size=size)
        points, normals = points[keep], normals[keep]
    else:
        raise ValueError("Supported methods are 'radius' and 'number'.")

    return Geometry(points, normals=normals, sampling_rate=geometry._sampling_rate)


@use_point_data()
def crop(geometry, distance: float, query: np.ndarray, keep_smaller: bool = True):
    """
    Filters points based on their distance to a set of query points.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    distance : float
        Distance threshold for cropping.
    query : np.ndarray
        Points to compute distances to.
    keep_smaller : bool, optional
        If True, keep points closer than distance threshold.
        If False, keep points farther than distance threshold.
        Default is True.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Cropped geometry.
    """
    dist = geometry.compute_distance(query_points=query, cutoff=distance)
    if keep_smaller:
        mask = dist < distance
    else:
        mask = dist >= distance

    return geometry[mask]


@use_point_data()
def sample(
    geometry, sampling: float, method: str, normal_offset: float = 0.0, **kwargs
):
    """
    Generates new points by sampling from a fitted parametric model.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    sampling : float
        Sampling rate or number of points to generate.
    method : str
        Sampling method to use. If not "N points", sampling is interpreted
        as a rate and converted to number of points.
    normal_offset : float, optional
        Point offset along normal vector, defaults to 0.0.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Sampled geometry.

    Raises
    ------
    ValueError
        If geometry has no fitted model.
    """
    from .geometry import Geometry

    if (fit := geometry.model) is None:
        return None

    n_samples, extra_kwargs = sampling, {}
    if method != "Points":
        n_samples = fit.points_per_sampling(sampling, normal_offset)
        extra_kwargs["mesh_init_factor"] = 5

    # We handle normal offset in sample to ensure equidistant spacing for meshes
    extra_kwargs["normal_offset"] = normal_offset
    points = fit.sample(int(n_samples), **extra_kwargs, **kwargs)
    normals = fit.compute_normal(points)

    return Geometry(points, normals=normals, sampling_rate=geometry._sampling_rate)


@use_point_data()
def trim(geometry, min_value: float, max_value: float, axis: str = "z"):
    """
    Filters points that fall within specified bounds along a coordinate axis.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    min_value : float
        Minimum bound value (inclusive).
    max_value : float
        Maximum bound value (inclusive).
    axis : str, optional
        Axis along which to trim ('x', 'y', or 'z').
        Default is 'z'.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Trimmed geometry.

    Raises
    ------
    ValueError
        If an invalid axis is provided.
    """
    _axis_map = {"x": 0, "y": 1, "z": 2}

    trim_column = _axis_map.get(axis.lower())
    if trim_column is None:
        raise ValueError(f"Axis must be one of {list(_axis_map.keys())}, got '{axis}'.")

    points = geometry.points
    coordinate_column = points[:, trim_column]
    mask = np.logical_and(
        coordinate_column >= min_value,
        coordinate_column <= max_value,
    )

    return geometry[mask]


@use_point_data(destructive=False)
def cluster(
    geometry,
    method: str,
    drop_noise: bool = False,
    use_points: bool = True,
    use_normals: bool = False,
    downsampling_radius: float = -1.0,
    **kwargs,
) -> List:
    """
    Partitions points into clusters using the specified clustering algorithm.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str
        Clustering method to use. Options are:
        - 'DBSCAN' : Density-based clustering
        - 'Birch' : Balanced iterative reducing clustering hierarchy
        - 'K-Means' : K-means clustering
        - 'Connected Components' : Connected component analysis
        - 'Envelope' : Envelope-based clustering
        - 'Leiden' : Leiden community detection
    drop_noise : bool, optional
        If True, drop noise points (label -1) from results.
        Default is False.
    use_points : bool, optional
        If True, use point coordinates for clustering.
        Default is True.
    use_normals : bool, optional
        If True, include normal vectors in clustering features.
        Default is False.
    downsampling_radius : float, optional
        Downsample point cloud based on radius and perform clustering. Subsequently
        all points are assigned to the nearest cluster in the downsampled set.
    **kwargs
        Additional arguments passed to the chosen clustering method.

    Returns
    -------
    List[:py:class:`mosaic.geometry.Geometry`]
        List of geometries, one per cluster.

    Raises
    ------
    ValueError
        If unsupported clustering method is specified or too many clusters found.
    """
    _mapping = {
        "DBSCAN": dbscan_clustering,
        "Birch": birch_clustering,
        "K-Means": kmeans_clustering,
        "Connected Components": connected_components,
        "Envelope": envelope_components,
        "Leiden": leiden_clustering,
    }
    func = _mapping.get(method)
    if func is None:
        raise ValueError(
            f"Method must be one of {list(_mapping.keys())}, got '{method}'."
        )

    points = geometry.points.copy()

    if downsampling_radius > 0:
        downsampled_geometry = downsample(
            geometry, method="radius", voxel_size=downsampling_radius
        )
        points = downsampled_geometry.points

    distance = geometry.sampling_rate
    if method in ("Connected Components", "Envelope", "Leiden"):
        distance = kwargs.pop("distance", -1)
        if np.any(np.array(distance) < 0):
            distance = geometry.sampling_rate
        kwargs["distance"] = distance

    distance = np.maximum(distance, downsampling_radius)
    points = np.divide(points, distance)

    # Prepare feature data for clustering
    data = points
    if use_points and use_normals:
        data = np.concatenate((points, geometry.normals), axis=1)
    elif not use_points and use_normals:
        data = geometry.normals

    labels = func(data, **kwargs)
    unique_labels = np.unique(labels)
    if len(unique_labels) > 10000:
        raise ValueError("Found more than 10k clusters. Try coarser clustering.")

    if downsampling_radius > 0:
        _, indices = find_closest_points(downsampled_geometry.points, geometry.points)
        labels = labels[indices]

    # Create geometry objects for each cluster
    result_geometries = []
    for label in unique_labels:
        if label == -1 and drop_noise:
            continue
        cluster_geometry = geometry[labels == label]
        result_geometries.append(cluster_geometry)

    return result_geometries


@use_point_data()
def remove_outliers(geometry, method: str = "statistical", **kwargs):
    """
    Filters out points that are statistical outliers based on local neighborhoods.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data.
    method : str, optional
        Outlier detection method. Options are:
        - 'statistical' : Statistical outlier removal based on neighbor distances
        - 'eigenvalue' : Eigenvalue-based outlier removal
        Default is 'statistical'.
    **kwargs
        Additional parameters for outlier removal method.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry` or None
        Filtered point cloud geometry with outliers removed.
        Returns None if no points remain after filtering.
    """
    func = statistical_outlier_removal
    if method == "eigenvalue":
        func = eigenvalue_outlier_removal
    else:
        if method != "statistical":
            raise ValueError(
                f"Unsupported method '{method}'. Use 'statistical' or 'eigenvalue'."
            )

    mask = func(geometry.points, **kwargs)
    if mask.sum() == 0:
        return None

    return geometry[mask]


@use_point_data()
def compute_normals(
    geometry, method: str = "Compute", k: int = 15, **kwargs
) -> Optional:
    """
    Calculates normals for points or flips existing normals.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Input data. This geometry object is modified in-place.
    method : str, optional
        Normal computation method. Options are:
        - 'Compute' : Calculate new normals from point neighborhoods
        - 'Flip' : Flip existing normals (multiply by -1)
        Default is 'Compute'.
    k : int, optional
        Number of neighbors to consider for normal computation.
        Only used when method='Compute'. Default is 15.
    **kwargs
        Additional parameters for normal computation.
    """
    from .utils import compute_normals

    if method == "Flip":
        geometry.normals = geometry.normals * -1
    elif method == "Compute":
        geometry.normals = compute_normals(geometry.points, k=k, **kwargs)
    else:
        raise ValueError(f"Unsupported method '{method}'. Use 'Compute' or 'Flip'.")
    return duplicate(geometry)


def duplicate(geometry, **kwargs):
    """
    Duplicate a geometry.

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Geometry to duplicate.

    Returns
    -------
    :py:class:`mosaic.geometry.Geometry`
        Duplicated geometry.
    """
    return geometry[...]


def visibility(geometry, visible: bool = True, **kwargs):
    """
    Change the visibility of a geometry object

    Parameters
    ----------
    geometry : :py:class:`mosaic.geometry.Geometry`
        Geometry to duplicate.
    visible: bool, optional
        Whether the Geometry instance should be visible or not.
    """
    geometry.set_visibility(visible)


class GeometryOperations:
    """Registry for geometry operation functions."""

    @classmethod
    def register(cls, operation_name: str, func, decorator=None):
        """Register an operation function."""
        if decorator is not None:
            func = decorator(func)
        setattr(cls, operation_name, staticmethod(func))


for operation_name, operation_func in [
    ("decimate", decimate),
    ("downsample", downsample),
    ("crop", crop),
    ("sample", sample),
    ("trim", trim),
    ("cluster", cluster),
    ("remove_outliers", remove_outliers),
    ("compute_normals", compute_normals),
    ("duplicate", duplicate),
    ("visibility", visibility),
]:
    GeometryOperations.register(operation_name, operation_func)



"""
Atomic Geometry class displayed by the vtk viewer.

Copyright (c) 2024-2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import warnings
from uuid import uuid4
from typing import Tuple, List, Dict

import vtk
import numpy as np
from vtk.util import numpy_support

from .actor import create_actor
from .utils import find_closest_points, normals_to_rot, apply_quat, NORMAL_REFERENCE


__all__ = ["Geometry", "VolumeGeometry", "GeometryTrajectory"]


BASE_COLOR = (0.7, 0.7, 0.7)


class Geometry:
    """
    VTK-based geometry representation for 3D point clouds and meshes.

    Parameters
    ----------
    points : np.ndarray, optional
        3D point coordinates.
    quaternions : np.ndarray, optional
        Normal vectors for each point (x,y,z).
    quaternions : np.ndarray, optional
        Orientation quaternions for each point (scalar first w,x,y,z).
    color : tuple, optional
        Base RGB color values, by default (0.7, 0.7, 0.7).
    sampling_rate : np.ndarray, optional
        Sampling rates along each axis.
    meta : dict, optional
        Metadata dictionary.
    vtk_actor : vtk.vtkActor, optional
        Custom VTK actor object.
    vertex_properties : VertexPropertyContainer, optional
        Additional vertex properties.
    model : :py:class:`mosaic.parametrization.Parametrization`
        Model fitted to geometry data.
    **kwargs
        Additional keyword arguments including normals.
    """

    def __init__(
        self,
        points=None,
        normals=None,
        quaternions=None,
        color=BASE_COLOR,
        sampling_rate=None,
        meta=None,
        vtk_actor=None,
        vertex_properties=None,
        model=None,
        **kwargs,
    ):
        self.uuid = str(uuid4())

        self._points = vtk.vtkPoints()
        self._points.SetDataTypeToFloat()

        self._cells = vtk.vtkCellArray()
        self._normals = vtk.vtkFloatArray()
        self._normals.SetNumberOfComponents(3)
        self._normals.SetName("Normals")

        self._data = vtk.vtkPolyData()
        self._data.SetPoints(self._points)
        self._data.SetVerts(self._cells)

        self.sampling_rate = sampling_rate

        if quaternions is not None:
            _normals = apply_quat(quaternions)
            if normals is not None:
                if not np.allclose(_normals, normals, atol=1e-3):
                    warnings.warn(
                        "Orientation given by quaternions does not match the "
                        "supplied normal vectors. Overwriting normals with "
                        "quaternions for now."
                    )
            normals = _normals

        if normals is None and points is not None:
            normals = np.full_like(points, fill_value=NORMAL_REFERENCE)

        if points is not None:
            self.points = points

        if normals is not None:
            self.normals = normals

        if quaternions is not None:
            self.quaternions = quaternions

        self._model = model
        self._cache = {}
        self._meta = {} if meta is None else meta
        self._representation = "pointcloud"

        self._actor = self._create_actor(vtk_actor)
        self._vertex_properties = vertex_properties
        self._appearance = {
            "size": 8,
            "opacity": 1.0,
            "ambient": 0.3,
            "diffuse": 0.7,
            "specular": 0.2,
            "render_spheres": True,
            "base_color": color,
        }
        self.set_appearance(**self._appearance)

    @property
    def model(self):
        return self._model

    @property
    def vertex_properties(self):
        return self._vertex_properties

    @property
    def sampling_rate(self):
        return np.asarray(self._sampling_rate).astype(np.float32)

    @sampling_rate.setter
    def sampling_rate(self, sampling_rate):
        if sampling_rate is None:
            sampling_rate = np.ones(3, dtype=np.float32)
        sampling_rate = np.asarray(sampling_rate, dtype=np.float32)
        sampling_rate = np.repeat(sampling_rate, 3 // sampling_rate.size)
        self._sampling_rate = sampling_rate

    def __getstate__(self):
        """
        Get object state for pickling.

        Returns
        -------
        dict
            Serializable state dictionary.
        """
        points, normals, quaternions = self.get_point_data()

        return {
            "points": points,
            "normals": normals,
            "quaternions": quaternions,
            "sampling_rate": self.sampling_rate,
            "meta": self._meta,
            "visible": self.visible,
            "appearance": self._appearance,
            "representation": self._representation,
            "vertex_properties": self.vertex_properties,
            "uuid": self.uuid,
            "model": self.model,
        }

    def __setstate__(self, state):
        """
        Restore object state from unpickling.

        Parameters
        ----------
        state : dict
            State dictionary to restore from.
        """
        uuid = state.pop("uuid", None)
        visible = state.pop("visible", True)
        appearance = state.pop("appearance", {})

        # Compatibility with pre 1.0.12
        if "fit" in state.get("meta", {}):
            state["model"] = state["meta"].pop("fit")

        self.__init__(**state)
        self.set_visibility(visible)

        if uuid is not None:
            self.uuid = uuid

        if "mesh" in state.get("cache", {}):
            self._cache["mesh"] = state["cache"]["mesh"]

        # Required to support loading VolumeGeometries
        if state.get("representation") != self._representation:
            self.change_representation(state.get("representation"))
        self.set_appearance(**appearance)

    def __getitem__(self, idx):
        """Array-like indexing using int/bool numpy arrays, slices or ellipsis."""
        if isinstance(idx, (int, np.integer)):
            idx = [idx]
        elif isinstance(idx, slice) or idx is ...:
            idx = np.arange(self.get_number_of_points())[idx]

        idx = np.asarray(idx)
        if idx.dtype == bool:
            idx = np.where(idx)[0]
        idx = idx[idx < self.get_number_of_points()]

        state = self.__getstate__()
        if (vertex_properties := state.get("vertex_properties")) is not None:
            state["vertex_properties"] = vertex_properties[idx]

        if "meta" in state:
            state["meta"] = state["meta"].copy()

        # Update underlying point data if available
        data_array = ("points", "normals", "quaternions")
        _point_data = self._cache.get("point_data")
        _is_surface = self._representation in ("mesh", "wireframe", "surface")
        if _is_surface and _point_data is not None:

            # Update mesh representatin in subset
            if "mesh" in self._cache:
                state["cache"] = {"mesh": self._cache["mesh"].subset(idx)}

            # Map points to closest vertex in the mesh (now in self.points)
            _, indices = find_closest_points(self.points, _point_data[0], k=1)

            # We dont need a copy in this case because the data is not modified
            idx = np.in1d(indices, idx)
            state |= {k: x for k, x in zip(data_array, _point_data)}

        for key in data_array:
            if (value := state.get(key)) is not None:
                state[key] = np.asarray(value)[idx].copy()

        _ = state.pop("uuid", None)
        ret = self.__class__.__new__(self.__class__)
        ret.__setstate__(state)
        return ret

    @classmethod
    def merge(cls, geometries):
        """
        Merge multiple geometry objects into a single geometry.

        Parameters
        ----------
        geometries : list of Geometry
            List of geometry objects to merge.

        Returns
        -------
        Geometry
            New geometry object containing merged data.

        Raises
        ------
        ValueError
            If no geometries provided for merging.
        """
        geometries = [x for x in geometries if isinstance(x, Geometry)]
        if not len(geometries):
            raise ValueError("No geometries provided for merging")
        elif len(geometries) == 1:
            return geometries[0]

        points, quat, normals, meshes = [], [], [], []
        for geometry in geometries:
            _points, _normals, _quaternions = geometry.get_point_data()

            points.append(_points)
            normals.append(_normals)
            if _quaternions is not None:
                quat.append(_quaternions)

            if (mesh := geometry._cache.get("mesh")) is not None:
                meshes.append(mesh)

        # Merging Geometries with different sampling rate is an underdetermined
        # problem without user intervention. Computing the maximum of geometries
        # makes the problem symmetric. In most workflows this should suffice, but
        # we might need to show a warning moving forward.
        sampling_rate = np.max(np.array([x.sampling_rate for x in geometries]), axis=0)

        # Use majority representation for new class
        representation = [x._representation for x in geometries]
        representation = max(set(representation), key=representation.count)
        appearance = [
            x._appearance for x in geometries if x._representation == representation
        ][0]

        cache = {}
        if len(meshes):
            from .parametrization import TriangularMesh
            from .meshing import merge_meshes, to_open3d

            vertices, faces = merge_meshes(
                vertices=[x.vertices for x in meshes],
                faces=[x.triangles for x in meshes],
            )
            cache["mesh"] = TriangularMesh(to_open3d(vertices, faces), repair=False)

        state = {
            "points": np.concatenate(points, axis=0),
            "normals": np.concatenate(normals, axis=0) if len(normals) else None,
            "quaternions": np.concatenate(quat, axis=0) if len(quat) else None,
            "sampling_rate": sampling_rate,
            "visible": any(x.visible for x in geometries),
            "representation": representation,
            "cache": cache,
            "appearance": appearance,
        }
        ret = cls.__new__(cls)
        ret.__setstate__(state)
        return ret

    @property
    def actor(self):
        """
        VTK actor object for rendering.

        Returns
        -------
        vtk.vtkActor
            VTK actor used for visualization.
        """
        return self._actor

    @property
    def visible(self):
        """
        Visibility state of the geometry.

        Returns
        -------
        bool
            True if geometry is visible, False otherwise.
        """
        return self.actor.GetVisibility()

    @property
    def points(self):
        """
        3D point coordinates of the geometry.

        Returns
        -------
        np.ndarray
            Point coordinates with shape (n_points, 3).
        """
        return numpy_support.vtk_to_numpy(self._data.GetPoints().GetData())

    @points.setter
    def points(self, points: np.ndarray):
        """
        Set 3D point coordinates.

        Parameters
        ----------
        points : np.ndarray
            Point coordinates with shape (n_points, 3).
        """
        points = np.asarray(points, dtype=np.float32)
        if points.shape[1] != 3:
            warnings.warn("Only 3D point clouds are supported.")
            return -1

        vertex_cells = vtk.vtkCellArray()
        idx = np.arange(points.shape[0], dtype=int)
        cells = np.column_stack((np.ones(idx.size, dtype=int), idx)).flatten()
        vertex_cells.SetCells(idx.size, numpy_support.numpy_to_vtkIdTypeArray(cells))

        self._points.SetData(numpy_support.numpy_to_vtk(points, deep=False))
        self._data.SetVerts(vertex_cells)
        self._data.Modified()

    @property
    def normals(self):
        """
        Normal vectors at each point.

        Returns
        -------
        np.ndarray or None
            Normal vectors with shape (n_points, 3), or None if not set.
        """
        normals = self._data.GetPointData().GetNormals()
        if normals is not None:
            normals = np.asarray(normals)
        return normals

    @normals.setter
    def normals(self, normals: np.ndarray):
        """
        Set normal vectors.

        Parameters
        ----------
        normals : np.ndarray
            Normal vectors with shape (n_points, 3).
        """
        normals = np.asarray(normals, dtype=np.float32)
        if normals.shape != self.points.shape:
            warnings.warn("Number of normals must match number of points.")
            return -1

        normals_vtk = numpy_support.numpy_to_vtk(normals, deep=True)
        normals_vtk.SetName("Normals")
        self._data.GetPointData().SetNormals(normals_vtk)

        # Update associated quaternions if available
        quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
        if quaternions is not None:
            self.quaternions = normals_to_rot(self.normals, scalar_first=True)
        self._data.Modified()

    @property
    def quaternions(self):
        """
        Orientation quaternions for each point.

        Returns
        -------
        np.ndarray or None
            Quaternions in scalar-first format (n_points, 4), or None if not set.
        """
        quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
        if quaternions is not None:
            quaternions = np.asarray(quaternions)
        elif self.normals is not None:
            warnings.warn("Computing quaternions from associated normals.")
            quaternions = normals_to_rot(self.normals, scalar_first=True)
            self.quaternions = quaternions
        return quaternions

    @quaternions.setter
    def quaternions(self, quaternions: np.ndarray):
        """
        Add orientation quaternions to the geometry.

        Parameters:
        -----------
        quaternions : array-like
            Quaternion values in scalar-first format (n, (w, x, y, z)).
        """
        quaternions = np.asarray(quaternions, dtype=np.float32)
        if quaternions.shape[0] != self.points.shape[0]:
            warnings.warn("Number of orientations must match number of points.")
            return -1
        if quaternions.shape[1] != 4:
            warnings.warn("Quaternions must have 4 components (w, x, y, z).")
            return -1

        quat_vtk = numpy_support.numpy_to_vtk(quaternions, deep=True)
        quat_vtk.SetName("OrientationQuaternion")
        self._data.GetPointData().AddArray(quat_vtk)
        self._data.Modified()

    def _set_faces(self, faces):
        faces = np.asarray(faces, dtype=int)
        if faces.shape[1] != 3:
            warnings.warn("Only triangular faces are supported.")
            return -1

        faces = np.concatenate(
            (np.full((faces.shape[0], 1), fill_value=3), faces), axis=1, dtype=int
        )
        poly_cells = vtk.vtkCellArray()
        poly_cells.SetCells(
            faces.shape[0], numpy_support.numpy_to_vtkIdTypeArray(faces.ravel())
        )
        self._data.SetPolys(poly_cells)
        self._data.Modified()

    def set_color(self, color: Tuple[int] = None):
        """
        Set uniform color for all points in the geometry.

        Parameters
        ----------
        color : tuple of int, optional
            RGB color values. Uses base color if None.
        """
        if color is None:
            color = self._appearance["base_color"]
        self.color_points(
            np.arange(self._points.GetNumberOfPoints(), dtype=np.int32), color=color
        )

    def set_visibility(self, visibility: bool = True):
        """
        Set geometry visibility in the scene.

        Parameters
        ----------
        visibility : bool, optional
            Whether geometry should be visible, by default True.
        """
        return self.actor.SetVisibility(visibility)

    def set_appearance(
        self,
        size: int = None,
        opacity: float = None,
        render_spheres: bool = None,
        ambient: float = None,
        diffuse: float = None,
        specular: float = None,
        color: Tuple[float] = None,
        **kwargs,
    ):
        """
        Set visual appearance properties of the geometry.

        Parameters
        ----------
        size : int, optional
            Point size for rendering.
        opacity : float, optional
            Transparency level (0.0 to 1.0).
        render_spheres : bool, optional
            Whether to render points as spheres.
        ambient : float, optional
            Ambient lighting coefficient.
        diffuse : float, optional
            Diffuse lighting coefficient.
        specular : float, optional
            Specular lighting coefficient.
        color : tuple of float, optional
            RGB color values.
        **kwargs
            Additional appearance parameters.
        """
        params = {
            "size": size,
            "opacity": opacity,
            "render_spheres": render_spheres,
            "ambient": ambient,
            "diffuse": diffuse,
            "specular": specular,
            **kwargs,
        }
        self._appearance.update({k: v for k, v in params.items() if v is not None})
        self._set_appearance()

        if color is None:
            color = self._appearance.get("base_color", (0.7, 0.7, 0.7))
        self.set_color(color)

    def _set_appearance(self):
        """Propagate appearance settings to VTK actor properties."""
        prop = self._actor.GetProperty()

        prop.SetRenderPointsAsSpheres(True)
        if not self._appearance.get("render_spheres", True):
            prop.SetRenderPointsAsSpheres(False)

        prop.SetPointSize(self._appearance.get("size", 8))
        prop.SetOpacity(self._appearance.get("opacity", 1.0))
        prop.SetAmbient(self._appearance.get("ambient", 0.3))
        prop.SetDiffuse(self._appearance.get("diffuse", 0.7))
        prop.SetSpecular(self._appearance.get("specular", 0.2))

    def _create_actor(
        self, actor=None, lod_points: int = 5e6, lod_points_size: int = 3
    ):
        """
        Create VTK actor with appropriate mapper configuration.

        Parameters
        ----------
        actor : vtk.vtkActor, optional
            Existing actor to use.
        lod_points : int, optional
            Level of detail threshold for points, by default 5e6.
        lod_points_size : int, optional
            Point size for level of detail, by default 3.

        Returns
        -------
        vtk.vtkActor
            Configured VTK actor.
        """
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(self._data)

        mapper.SetScalarModeToDefault()
        mapper.SetVBOShiftScaleMethod(1)
        mapper.SetResolveCoincidentTopology(False)
        mapper.SetResolveCoincidentTopologyToPolygonOffset()

        if actor is None:
            actor = create_actor()
        actor.SetMapper(mapper)
        return actor

    def get_number_of_points(self):
        """
        Get total number of points in the geometry.

        Returns
        -------
        int
            Number of points.
        """
        return self._points.GetNumberOfPoints()

    def set_scalars(self, scalars, color_lut, scalar_range=None, use_point=False):
        """
        Set scalar data for coloring points.

        Parameters
        ----------
        scalars : array-like
            Scalar values for each point.
        color_lut : vtk.vtkLookupTable
            Color lookup table for mapping scalars to colors.
        scalar_range : tuple, optional
            Min and max scalar range for color mapping.
        use_point : bool, optional
            Whether to use point data for scalar mode, by default False.
        """
        scalars = np.asarray(scalars).ravel()
        if scalars.size == 1:
            scalars = np.full(
                (self.get_number_of_points()), fill_value=scalars, dtype=scalars.dtype
            )

        if scalars.size != self.points.shape[0]:
            return None

        mapper = self._actor.GetMapper()
        mapper.GetInput().GetPointData().SetScalars(
            numpy_support.numpy_to_vtk(scalars, deep=False)
        )

        self._configure_scalar_mapper(mapper, color_lut, scalar_range, use_point)
        self._actor.Modified()

    def _update_scalars_from_ids(
        self, point_ids, color_lut, scalar_range=None, use_point=False
    ) -> bool:
        """
        Try to update existing scalar array in place for better performance.

        Parameters
        ----------
        point_ids : array-like
            Point indices to set to 1.0, all others set to 0.0
        color_lut : vtk.vtkLookupTable
            Color lookup table for mapping scalars to colors.
        scalar_range : tuple, optional
            Min and max scalar range for color mapping.
        use_point : bool, optional
            Whether to use point data for scalar mode, by default False.

        Returns
        -------
        bool
            True if successful, False if scalar array couldn't be reused.
        """
        mapper = self._actor.GetMapper()
        cur_scalars = mapper.GetInput().GetPointData().GetScalars()

        if not (cur_scalars is not None and cur_scalars.GetNumberOfComponents() == 1):
            return False

        scalars_np = vtk.util.numpy_support.vtk_to_numpy(cur_scalars)
        if scalars_np.shape[0] != self._points.GetNumberOfPoints():
            return False

        scalars_np.fill(0.0)
        scalars_np[point_ids] = 1.0
        cur_scalars.Modified()

        self._configure_scalar_mapper(mapper, color_lut, scalar_range, use_point)
        self._actor.Modified()
        return True

    def _configure_scalar_mapper(
        self, mapper, color_lut, scalar_range=None, use_point=False
    ):
        """
        Configure mapper for scalar coloring with common settings.

        Parameters
        ----------
        mapper : vtk.vtkMapper
            The mapper to configure
        color_lut : vtk.vtkLookupTable
            Color lookup table for mapping scalars to colors.
        scalar_range : tuple, optional
            Min and max scalar range for color mapping.
        use_point : bool, optional
            Whether to use point data for scalar mode, by default False.
        """
        if color_lut is not None:
            mapper.SetLookupTable(color_lut)
        if scalar_range is not None:
            mapper.SetScalarRange(*scalar_range)
        mapper.ScalarVisibilityOn()
        if use_point:
            mapper.SetScalarModeToUsePointData()

    def color_points(self, point_ids: set, color: Tuple[float]):
        """
        Color specific points in the geometry using set_scalars backend.

        Parameters
        ----------
        point_ids : np.ndarray
            Set of point indices to color
        color : tuple of float
            RGB color values (0-1) to apply to selected points
        """
        n_points = self._points.GetNumberOfPoints()
        if not isinstance(point_ids, np.ndarray):
            point_ids = np.asarray(point_ids, dtype=np.int32, copy=False)

        point_ids = point_ids.astype(np.int32, copy=False)
        point_ids = point_ids[point_ids < n_points]

        lut = vtk.vtkLookupTable()
        lut.SetNumberOfTableValues(2)
        lut.SetRange(0.0, 1.0)
        lut.SetTableValue(0, *self._appearance["base_color"], 1.0)
        lut.SetTableValue(1, *color, 1.0)
        lut.Build()

        kw = {"color_lut": lut, "scalar_range": (0.0, 1.0), "use_point": True}

        success = self._update_scalars_from_ids(point_ids, **kw)
        if not success:
            scalars = np.zeros(n_points, dtype=np.float32)
            scalars[point_ids] = 1.0
            return self.set_scalars(scalars, **kw)

    def swap_data(
        self,
        points,
        normals=None,
        faces=None,
        quaternions=None,
        model=None,
        meta: Dict = None,
    ):
        """
        Replace geometry data with new point cloud or mesh data.

        Parameters
        ----------
        points : np.ndarray
            New 3D point coordinates.
        normals : np.ndarray, optional
            New normal vectors.
        faces : np.ndarray, optional
            New face connectivity indices.
        quaternions : np.ndarray, optional
            New orientation quaternions.
        model : :py:class:`mosaic.parametrization.Parametrization`
            Model fitted to geometry data.
        meta : dict, optional
            New metadata dictionary.

        Returns
        -------
        int
            Result of representation change.
        """
        self._points.Reset()
        self._cells.Reset()
        self._normals.Reset()

        # Check whether we have to synchronize quaternion representation
        _quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
        if quaternions is None and _quaternions is not None and normals is not None:
            quaternions = normals_to_rot(normals)

        self.points = points
        if quaternions is not None:
            normals = apply_quat(quaternions, NORMAL_REFERENCE)
            self.quaternions = quaternions

        if normals is None and points is not None:
            normals = np.full_like(points, fill_value=NORMAL_REFERENCE)

        if normals is not None:
            self.normals = normals

        if faces is not None:
            self._set_faces(faces)

        self._model = model
        if isinstance(meta, dict):
            self._meta.update(meta)

            if "vertex_properties" in meta:
                self._vertex_properties = meta["vertex_properties"]

        self.set_color()
        return self.change_representation(self._representation)

    def change_representation(self, representation: str = "pointcloud") -> int:
        """
        Change the visual representation mode of the geometry.

        Parameters
        ----------
        representation : str, optional
            Representation mode, by default "pointcloud".
            Supported: "pointcloud", "gaussian_density", "pointcloud_normals",
            "mesh", "wireframe", "normals", "surface", "basis".

        Returns
        -------
        int
            Success status (0 for success, -1 for failure).

        Raises
        ------
        ValueError
            If representation mode is not supported.
        """
        supported = [
            "pointcloud",
            "gaussian_density",
            "mesh",
            "wireframe",
            "normals",
            "surface",
            "basis",
        ]
        representation = representation.lower()

        # We dont check representation == self._representation to enable
        # rendering in the same representation after swap_data
        if representation not in supported:
            supported = ", ".join(supported)
            raise ValueError(
                f"Supported representations are {supported} - got {representation}."
            )
        clipping_planes = self._actor.GetMapper().GetClippingPlanes()

        # Use fitted mesh representation or create a new one
        to_mesh = self.is_mesh_representation(representation)
        if to_mesh:
            mesh = self.model
            if mesh is None and "mesh" in self._cache:
                mesh = self._cache["mesh"]

            if not hasattr(mesh, "mesh"):
                from .parametrization import FlyingEdges

                try:
                    mesh = FlyingEdges.fit(self.points, np.max(self.sampling_rate))
                except Exception as e:
                    warnings.warn(f"Failed to mesh object: {str(e)}.")
                    return None
                self._cache["mesh"] = mesh

        mapper = vtk.vtkPolyDataMapper()
        if representation == "gaussian_density":
            mapper = vtk.vtkPointGaussianMapper()
            mapper.SetSplatShaderCode("")

        mapper.SetScalarModeToDefault()
        mapper.SetVBOShiftScaleMethod(1)
        mapper.SetResolveCoincidentTopology(False)
        mapper.SetResolveCoincidentTopologyToPolygonOffset()

        self._actor.SetMapper(mapper)
        self._appearance.update({"opacity": 1, "size": 8, "render_spheres": True})
        if representation == "gaussian_density":
            self._appearance["render_spheres"] = False

        # Backup to support treating point clouds as mesh
        if self._cache.get("point_data") is not None and not to_mesh:
            self._points = vtk.vtkPoints()
            self._points.SetDataTypeToFloat()

            self._cells = vtk.vtkCellArray()
            self._normals = vtk.vtkFloatArray()
            self._normals.SetNumberOfComponents(3)
            self._normals.SetName("Normals")

            self._data = vtk.vtkPolyData()
            self._data.SetPoints(self._points)
            self._data.SetVerts(self._cells)

            self.points, self.normals, quaternions = self.get_point_data()
            if quaternions is not None:
                self.quaternions = quaternions
            self._cache["point_data"] = None

        scale = 15 * np.max(self.sampling_rate)
        mapper, prop = self._actor.GetMapper(), self._actor.GetProperty()
        prop.SetOpacity(self._appearance["opacity"])
        prop.SetPointSize(self._appearance["size"])
        prop.SetRenderPointsAsSpheres(self._appearance["render_spheres"])
        if representation == "pointcloud":
            prop.SetRepresentationToPoints()
            mapper.SetInputData(self._data)

        elif representation == "gaussian_density":
            mapper.SetSplatShaderCode("")
            mapper.SetScaleFactor(self._appearance["size"] * 0.25)
            mapper.SetScalarVisibility(True)
            mapper.SetInputData(self._data)

        elif representation == "normals":
            arrow = vtk.vtkArrowSource()
            arrow.SetTipResolution(6)
            arrow.SetShaftResolution(6)
            arrow.SetTipRadius(0.08)
            arrow.SetShaftRadius(0.02)

            mapper = vtk.vtkGlyph3DMapper()
            mapper.SetInputData(self._data)
            mapper.SetSourceConnection(arrow.GetOutputPort())
            mapper.SetOrientationArray("Normals")
            mapper.SetOrientationModeToDirection()
            mapper.SetScaleModeToNoDataScaling()
            mapper.SetScaleFactor(scale)
            mapper.OrientOn()

            self._actor.SetMapper(mapper)

        elif representation == "basis":
            if self.quaternions is None:
                print("Quaternions are required for basis representation.")
                return -1

            arrow_x = vtk.vtkArrowSource()
            arrow_y = vtk.vtkArrowSource()
            arrow_z = vtk.vtkArrowSource()
            for arrow in [arrow_x, arrow_y, arrow_z]:
                arrow.SetTipResolution(6)
                arrow.SetShaftResolution(6)
                arrow.SetTipRadius(0.08)
                arrow.SetShaftRadius(0.02)

            transform_x = vtk.vtkTransform()
            transform_x.RotateY(-90)
            transform_filter_x = vtk.vtkTransformPolyDataFilter()
            transform_filter_x.SetInputConnection(arrow_x.GetOutputPort())
            transform_filter_x.SetTransform(transform_x)
            transform_filter_x.Update()

            transform_y = vtk.vtkTransform()
            transform_y.RotateZ(90)
            transform_filter_y = vtk.vtkTransformPolyDataFilter()
            transform_filter_y.SetInputConnection(arrow_y.GetOutputPort())
            transform_filter_y.SetTransform(transform_y)
            transform_filter_y.Update()

            append_filter = vtk.vtkAppendPolyData()
            append_filter.AddInputConnection(transform_filter_x.GetOutputPort())
            append_filter.AddInputConnection(transform_filter_y.GetOutputPort())
            append_filter.AddInputConnection(arrow_z.GetOutputPort())
            append_filter.Update()

            mapper = vtk.vtkGlyph3DMapper()
            mapper.SetInputData(self._data)
            mapper.SetSourceData(append_filter.GetOutput())
            mapper.SetOrientationArray("OrientationQuaternion")
            mapper.SetOrientationModeToQuaternion()
            mapper.SetScaleFactor(scale)
            mapper.SetScaleModeToNoDataScaling()
            mapper.OrientOn()

            self._actor.SetMapper(mapper)

        elif to_mesh:
            # Avoid computing quaternions
            quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
            if quaternions is not None:
                quaternions = self.quaternions

            # No backup if we are dealing with a fitted mesh
            if not np.array_equal(mesh.vertices, self.points):
                self._cache["point_data"] = self.get_point_data()

            self._cells.Reset()
            self._points.Reset()

            self.points = mesh.vertices
            self._set_faces(mesh.triangles)
            self.normals = mesh.compute_vertex_normals()

            if representation in ("surface", "wireframe"):
                self._data.SetVerts(None)

            mapper.SetInputData(self._data)
            if representation == "wireframe":
                prop.SetRepresentationToWireframe()
            else:
                prop.SetRepresentationToSurface()
                prop.SetEdgeVisibility(representation == "mesh")

                self._appearance["size"] = 2
                prop.SetPointSize(self._appearance["size"])

        if clipping_planes:
            mapper.SetClippingPlanes(clipping_planes)

        self._representation = representation
        return self.set_appearance()

    def compute_distance(self, query_points: np.ndarray, k: int = 1, **kwargs):
        if hasattr(self.model, "compute_distance"):
            return self.model.compute_distance(query_points)

        return find_closest_points(self.points, query_points, k=k)[0]

    def is_mesh_representation(self, representation: str = None) -> bool:
        if representation is None:
            representation = self._representation
        return representation in ("mesh", "surface", "wireframe")

    def get_point_data(self):
        if (point_data := self._cache.get("point_data")) is None:
            quaternions = self._data.GetPointData().GetArray("OrientationQuaternion")
            if quaternions is not None:
                quaternions = self.quaternions
            return self.points, self.normals, quaternions
        return point_data


# For backwards compatibility
class PointCloud(Geometry):
    pass


class VolumeGeometry(Geometry):
    """
    Geometry class specialized for 3D volume rendering with isosurfaces.

    Parameters
    ----------
    volume : np.ndarray, optional
        3D volume data array.
    volume_sampling_rate : np.ndarray, optional
        Sampling rates for volume data, by default ones(3).
    **kwargs
        Additional keyword arguments passed to parent Geometry class.
    """

    def __init__(
        self, volume: np.ndarray = None, volume_sampling_rate=np.ones(3), **kwargs
    ):
        super().__init__(**kwargs)
        self._volume = None
        if volume is None:
            return None

        self._volume = vtk.vtkImageData()
        self._volume.SetSpacing(volume_sampling_rate)
        self._volume.SetDimensions(volume.shape)
        self._volume.AllocateScalars(vtk.VTK_FLOAT, 1)

        if self.quaternions is None:
            self.quaternions = normals_to_rot(self.normals, scalar_first=True)

        self._raw_volume = volume
        volume_vtk = numpy_support.numpy_to_vtk(
            volume.ravel(order="F"), deep=True, array_type=vtk.VTK_FLOAT
        )
        self._volume.GetPointData().SetScalars(volume_vtk)

        bounds = [0.0] * 6
        self._volume.GetBounds(bounds)
        transform = vtk.vtkTransform()
        transform.Translate(
            [-(b[1] - b[0]) * 0.5 for b in zip(bounds[::2], bounds[1::2])]
        )

        self._volume_sampling_rate = volume_sampling_rate

        # Render volume isosurface as vtk glpyh object
        transformFilter = vtk.vtkTransformFilter()
        transformFilter.SetInputData(self._volume)
        transformFilter.SetTransform(transform)
        transformFilter.Update()
        self._surface = vtk.vtkContourFilter()
        self._surface.SetInputConnection(transformFilter.GetOutputPort())
        self._surface.GenerateValues(1, volume.min(), volume.max())

        mapper = vtk.vtkGlyph3DMapper()
        mapper.SetInputData(self._data)
        mapper.SetSourceConnection(self._surface.GetOutputPort())
        mapper.SetOrientationModeToQuaternion()
        mapper.SetScaleModeToNoDataScaling()
        mapper.SetOrientationArray("OrientationQuaternion")
        mapper.OrientOn()
        self._actor.SetMapper(mapper)

        if "upper_quantile" in kwargs and "lower_quantile" in kwargs:
            self.update_isovalue_quantile(
                upper_quantile=kwargs.get("upper_quantile"),
                lower_quantile=kwargs.get("lower_quantile"),
            )
        self._representation = "volume"

    def __getstate__(self):
        state = super().__getstate__()

        if self._volume is not None:
            state.update(
                {
                    "volume": self._raw_volume,
                    "volume_sampling_rate": self._volume_sampling_rate,
                    "lower_quantile": self._lower_quantile,
                    "upper_quantile": self._upper_quantile,
                }
            )
        return state

    def update_isovalue(self, upper, lower: float = 0):
        """
        Update the isovalue for volume surface rendering.

        Parameters
        ----------
        upper : float
            Upper isovalue threshold.
        lower : float, optional
            Lower isovalue threshold, by default 0.
        """
        return self._surface.SetValue(int(lower), upper)

    def update_isovalue_quantile(
        self, upper_quantile: float, lower_quantile: float = 0.0
    ):
        """
        Update isovalue using quantile-based thresholds.

        Parameters
        ----------
        upper_quantile : float
            Upper quantile threshold (0.0 to 1.0).
        lower_quantile : float, optional
            Lower quantile threshold (0.0 to 1.0), by default 0.0.

        Raises
        ------
        ValueError
            If upper quantile is not greater than lower quantile.
        """
        lower_quantile = max(lower_quantile, 0)
        upper_quantile = min(upper_quantile, 1)

        if lower_quantile >= upper_quantile:
            raise ValueError("Upper quantile must be greater than lower quantile")

        self._lower_quantile = lower_quantile
        self._upper_quantile = upper_quantile
        lower_value = np.quantile(self._raw_volume, self._lower_quantile)
        upper_value = np.quantile(self._raw_volume, self._upper_quantile)
        return self.update_isovalue(upper=upper_value, lower=lower_value)

    def set_appearance(self, isovalue_percentile=99.5, **kwargs):
        if hasattr(self, "_raw_volume"):
            self._appearance["isovalue_percentile"] = isovalue_percentile
            self.update_isovalue_quantile(upper_quantile=isovalue_percentile / 100)
        super().set_appearance(**kwargs)


class GeometryTrajectory(Geometry):
    """
    Geometry class for displaying animated trajectory sequences.

    Parameters
    ----------
    trajectory : list of dict
        List of trajectory frames containing geometry data.
    **kwargs
        Additional keyword arguments passed to parent Geometry class.
    """

    def __init__(self, trajectory: List[Dict], **kwargs):
        super().__init__(**kwargs)
        self._trajectory = trajectory

    def __getstate__(self):
        state = super().__getstate__()
        state.update({"trajectory": self._trajectory})
        return state

    @property
    def frames(self):
        """
        Number of frames in the trajectory.

        Returns
        -------
        int
            Total number of trajectory frames.
        """
        return len(self._trajectory)

    def display_frame(self, frame_idx: int) -> bool:
        """
        Display specific trajectory frame.

        Parameters
        ----------
        frame_idx : int
            Index of frame to display.

        Returns
        -------
        bool
            True if frame was successfully displayed, False otherwise.
        """
        if frame_idx < 0 or frame_idx > self.frames:
            return False

        appearance = self._appearance.copy()

        meta = self._trajectory[frame_idx]
        model = meta.get("fit")
        if not hasattr(model, "mesh"):
            return False

        meta = {k: v for k, v in meta.items() if k != "fit"}
        self.swap_data(
            points=model.vertices,
            faces=model.triangles,
            normals=model.compute_vertex_normals(),
            meta=meta,
            model=model,
        )
        self.set_appearance(**appearance)
        return True



from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QWidget,
    QFrame,
    QFileDialog,
)


class PathSelector(QWidget):
    """Reusable component for file path selection with browse button"""

    def __init__(
        self,
        label_text="",
        placeholder="Path to file",
        file_mode: bool = True,
        parent=None,
    ):
        """
        Initialize the file path selector widget.

        Parameters:
        -----------
        label_text : str
            Text to show as label above the path field
        placeholder : str
            Placeholder text for the input field
        file_mode : str
            Whether integrated button triggers file or directory selection.
        parent : QWidget
            Parent widget
        """
        super().__init__(parent)

        self.file_mode = file_mode
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(4)

        if label_text:
            self.label = QLabel(label_text)
            self.label.setStyleSheet(
                """
                QLabel {
                    font-size: 12px;
                    font-weight: 500;
                    margin-bottom: 1px;
                }
            """
            )
            main_layout.addWidget(self.label)

        self.container_frame = QFrame()
        self.container_frame.setObjectName("container_frame")
        self.container_frame.setStyleSheet(
            """
            #container_frame {
                border: 1px solid #d1d5db;
                border-radius: 4px;
            }
            #container_frame:focus-within {
                border-color: #6366f1;
            }
        """
        )

        container_layout = QHBoxLayout(self.container_frame)
        container_layout.setContentsMargins(0, 0, 0, 0)
        container_layout.setSpacing(0)

        self.path_input = QLineEdit()
        self.path_input.setPlaceholderText(placeholder)
        self.path_input.setStyleSheet(
            """
            QLineEdit {
                border: none;
                background-color: transparent;
                padding: 6px 8px;
            }
        """
        )

        self.browse_button = QPushButton("Browse")
        self.browse_button.setStyleSheet(
            """
            QPushButton {
                background-color: #f3f4f6;
                border: none;
                border-left: 1px solid #d1d5db;
                padding: 6px 12px;
                min-width: 70px;
                color: #374151;
            }
            QPushButton:hover {
                background-color: #e5e7eb;
            }
            QPushButton:pressed {
                background-color: #d1d5db;
            }
        """
        )
        self.browse_button.clicked.connect(self._browse_clicked)

        container_layout.addWidget(self.path_input, 1)
        container_layout.addWidget(self.browse_button)

        main_layout.addWidget(self.container_frame)

    def _browse_clicked(self):
        if self.file_mode:
            path, _ = QFileDialog.getOpenFileName(self, "Select File")
        else:
            path = QFileDialog.getExistingDirectory(self, "Select Directory")

        if path:
            return self.set_path(path)

    def get_path(self):
        """Get the currently entered path"""
        return self.path_input.text()

    def set_path(self, path):
        """Set the path in the input field"""
        self.path_input.setText(path)



from qtpy.QtCore import Qt, QRect, Signal
from qtpy.QtWidgets import QWidget, QScrollArea, QApplication
from qtpy.QtGui import QPainter, QColor, QPen, QFont, QMouseEvent

import qtawesome as qta


class TimelineContent(QWidget):
    trackSelected = Signal(str)
    frameMoved = Signal(int)
    trackMoved = Signal(str, int)
    trackRemoved = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.tracks = []
        self.current_frame = 0
        self.selected_track = None
        self.dragging_track = None
        self.drag_start_x = 0
        self.drag_start_frame = 0

        self.view_start_frame = 0
        self.view_end_frame = 300
        self.min_visible_frames = 1

        self.track_height = 40
        self.ruler_height = 30
        self.remove_button_size = 16
        self.hovered_remove_button = None

        self.setMouseTracking(True)
        self.update_size()

    def update_size(self):
        min_height = self.ruler_height + 10
        if self.tracks:
            min_height += len(self.tracks) * self.track_height

        self.setMinimumHeight(max(100, min_height))

    def set_tracks(self, tracks):
        self.tracks = tracks
        self.update_view_range()
        self.update_size()
        self.update()

    def set_current_frame(self, frame):
        self.current_frame = frame

        if frame < self.view_start_frame:
            self.view_start_frame = frame
            self.view_end_frame = frame + (self.view_end_frame - self.view_start_frame)
        elif frame > self.view_end_frame:
            self.view_end_frame = frame
            self.view_start_frame = frame - (
                self.view_end_frame - self.view_start_frame
            )

        self.update()

    def update_view_range(self):
        """Update view range based on track content"""
        if not self.tracks:
            return None

        last_frame = 0
        for track in self.tracks:
            track_end = track.animation.global_start_frame + track.animation.duration
            last_frame = max(last_frame, track_end)

        self.view_end_frame = max(self.min_visible_frames, last_frame + 20)

        if self.view_start_frame > self.view_end_frame - self.min_visible_frames:
            self.view_start_frame = max(
                0, self.view_end_frame - self.min_visible_frames
            )

    def frame_to_x(self, frame):
        """Convert frame number to x coordinate"""
        view_range = self.view_end_frame - self.view_start_frame
        if view_range == 0:
            return 0
        return (frame - self.view_start_frame) / view_range * self.width()

    def x_to_frame(self, x):
        """Convert x coordinate to frame number"""
        view_range = self.view_end_frame - self.view_start_frame
        frame = self.view_start_frame + (x / self.width()) * view_range
        return int(round(frame))

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        font = QFont(painter.font())
        font.setPointSize(8)
        painter.setFont(font)

        painter.setPen(QPen(Qt.GlobalColor.black, 1))
        painter.drawLine(0, self.ruler_height, self.width(), self.ruler_height)

        view_range = self.view_end_frame - self.view_start_frame
        pixels_per_frame = self.width() / view_range

        target_spacing = 80
        frames_per_tick = max(1, int(target_spacing / pixels_per_frame))

        # Find the smallest interval that's larger than frames_per_tick
        intervals = [1, 2, 5, 10, 20, 50, 100, 200, 500, 1000]
        for interval in intervals:
            if interval >= frames_per_tick:
                frames_per_tick = interval
                break
        else:
            frames_per_tick = 1000 * ((frames_per_tick + 500) // 1000)

        first_visible_tick = (
            int(self.view_start_frame) // frames_per_tick
        ) * frames_per_tick
        if first_visible_tick < self.view_start_frame:
            first_visible_tick += frames_per_tick

        metadata_font = QFont(painter.font())
        metadata_font.setPointSize(12)
        painter.setFont(metadata_font)

        for frame in range(
            first_visible_tick,
            int(self.view_end_frame) + frames_per_tick,
            frames_per_tick,
        ):
            x = self.frame_to_x(frame)
            if x < 0 or x > self.width():
                continue

            painter.drawLine(int(x), self.ruler_height - 5, int(x), self.ruler_height)

            label = str(frame)
            text_width = painter.fontMetrics().horizontalAdvance(label)
            text_x = x - text_width / 2
            if text_x > 0 and text_x + text_width < self.width() - 5:
                painter.drawText(
                    int(text_x), 5, text_width, 20, Qt.AlignmentFlag.AlignCenter, label
                )

        track_y = self.ruler_height + 10
        for track in self.tracks:
            start_x = self.frame_to_x(track.animation.global_start_frame)
            end_x = self.frame_to_x(
                track.animation.global_start_frame + track.animation.duration
            )

            x = max(0, start_x)
            width = min(self.width(), end_x) - x
            if end_x < 0 or start_x > self.width() or width <= 0:
                track_y += self.track_height
                continue

            color = QColor(track.color)
            if not track.animation.enabled:
                color.setAlpha(128)
            painter.fillRect(int(x), track_y, int(width), self.track_height - 10, color)

            painter.setPen(QPen(Qt.GlobalColor.black, 1))
            if track.id == self.selected_track:
                painter.setPen(QPen(Qt.GlobalColor.blue, 2))

            painter.drawRect(int(x), track_y, int(width), self.track_height - 10)
            if width > 40:
                painter.setPen(Qt.GlobalColor.white)
                text_rect = QRect(
                    int(x + 5), track_y + 5, int(width - 10), self.track_height - 20
                )
                painter.drawText(
                    text_rect,
                    Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
                    track.animation.name,
                )

            # Only draw if visible
            remove_x = int(end_x - self.remove_button_size - 5)
            remove_y = track_y + (self.track_height - 10 - self.remove_button_size) // 2
            if start_x < remove_x < self.width():
                icon_color = (
                    QColor(255, 255, 255)
                    if self.hovered_remove_button == track.id
                    else QColor(128, 128, 128)
                )
                icon = qta.icon("fa5s.trash", color=icon_color)
                pixmap = icon.pixmap(self.remove_button_size, self.remove_button_size)
                painter.drawPixmap(remove_x, remove_y, pixmap)

            track_y += self.track_height

        playhead_x = self.frame_to_x(self.current_frame)
        if 0 <= playhead_x <= self.width():
            painter.setPen(QPen(Qt.GlobalColor.red, 1))
            painter.drawLine(int(playhead_x), 0, int(playhead_x), self.height())

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() != Qt.MouseButton.LeftButton or not self.tracks:
            return super().mousePressEvent(event)

        track_y = self.ruler_height + 10
        for track in self.tracks:
            start_x = self.frame_to_x(track.animation.global_start_frame)
            end_x = self.frame_to_x(
                track.animation.global_start_frame + track.animation.duration
            )

            # Check if click is on remove button
            remove_x = int(end_x - self.remove_button_size - 5)
            remove_y = track_y + (self.track_height - 10 - self.remove_button_size) // 2

            if (
                remove_x <= event.x() <= remove_x + self.remove_button_size
                and remove_y <= event.y() <= remove_y + self.remove_button_size
                and start_x < remove_x < self.width()
            ):
                self.trackRemoved.emit(track.id)

            if (
                start_x <= event.x() <= end_x
                and track_y <= event.y() <= track_y + self.track_height - 10
            ):
                self.selected_track = track.id
                self.dragging_track = track.id
                self.drag_start_x = event.x()
                self.drag_start_frame = track.animation.global_start_frame
                self.trackSelected.emit(track.id)
                self.update()
                return

            track_y += self.track_height

        frame = self.x_to_frame(event.x())
        self.current_frame = max(0, frame)
        self.frameMoved.emit(self.current_frame)
        self.update()

    def mouseMoveEvent(self, event: QMouseEvent):
        # Check if we're hovering over a remove button
        if not (event.buttons() & Qt.MouseButton.LeftButton):
            hovered_button = None
            track_y = self.ruler_height + 10
            for track in self.tracks:
                start_x = self.frame_to_x(track.animation.global_start_frame)
                end_x = self.frame_to_x(
                    track.animation.global_start_frame + track.animation.duration
                )

                remove_x = int(end_x - self.remove_button_size - 5)
                remove_y = (
                    track_y + (self.track_height - 10 - self.remove_button_size) // 2
                )

                if (
                    remove_x <= event.x() <= remove_x + self.remove_button_size
                    and remove_y <= event.y() <= remove_y + self.remove_button_size
                    and start_x < remove_x < self.width()
                ):
                    hovered_button = track.id
                    break

                track_y += self.track_height

            if hovered_button != self.hovered_remove_button:
                self.hovered_remove_button = hovered_button
                self.update()

        draggable = self.dragging_track and self.tracks
        if not draggable or not (event.buttons() & Qt.MouseButton.LeftButton):
            return super().mouseMoveEvent(event)

        frame = self.x_to_frame(event.x())
        frame_shift = frame - self.x_to_frame(self.drag_start_x)
        new_frame = max(0, self.drag_start_frame + frame_shift)

        self.trackMoved.emit(self.dragging_track, new_frame)

        track = next((t for t in self.tracks if t.id == self.dragging_track), None)
        if track is None:
            return None

        track_end = new_frame + track.animation.duration
        if track_end > self.view_end_frame - 10:
            self.view_end_frame = track_end + 20
            self.update()

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.dragging_track = False
        return self.update()

    def wheelEvent(self, event):
        if not self.tracks:
            return super().wheelEvent(event)

        if not (event.modifiers() & Qt.ControlModifier):
            return event.ignore()

        mouse_frame = self.x_to_frame(event.position().x())
        view_range = self.view_end_frame - self.view_start_frame

        zoom_factor = 1.05
        new_range = max(self.min_visible_frames, view_range / zoom_factor)
        if event.angleDelta().y() <= 0:
            new_range = view_range * zoom_factor
            last_frame = 1
            for track in self.tracks:
                track_end = (
                    track.animation.global_start_frame + track.animation.duration
                )
                last_frame = max(last_frame, track_end)
            max_range = max(self.min_visible_frames, last_frame + 40)
            new_range = min(new_range, max_range)

        mouse_ratio = (mouse_frame - self.view_start_frame) / view_range
        self.view_start_frame = mouse_frame - new_range * mouse_ratio
        self.view_end_frame = self.view_start_frame + new_range

        if self.view_start_frame < 0:
            self.view_start_frame = 0
            self.view_end_frame = new_range

        self.update()


class TimelineWidget(QScrollArea):
    trackSelected = Signal(str)
    frameMoved = Signal(int)
    trackMoved = Signal(str, int)
    trackRemoved = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.content = TimelineContent(self)

        self.content.trackSelected.connect(self.trackSelected.emit)
        self.content.frameMoved.connect(self.frameMoved.emit)
        self.content.trackMoved.connect(self.trackMoved.emit)
        self.content.trackRemoved.connect(self.trackRemoved.emit)

        self.setWidget(self.content)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setFrameShape(self.NoFrame)

    def set_tracks(self, tracks):
        self.content.set_tracks(tracks)
        QApplication.processEvents()

    def set_current_frame(self, frame):
        self.content.set_current_frame(frame)
        QApplication.processEvents()

    @property
    def selected_track(self):
        return self.content.selected_track

    @selected_track.setter
    def selected_track(self, value):
        self.content.selected_track = value

    def update(self):
        self.content.update_view_range()
        self.content.update_size()
        self.content.update()
        QApplication.processEvents()
